#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import numbers

import numpy as np
import pytest
from helper import get_queue_or_skip, skip_if_dtype_not_supported

import dpctl
import dpctl.memory as dpm
import dpctl.tensor as dpt
from dpctl.tensor import Device


@pytest.mark.parametrize(
    "shape",
    [
        (),
        (4,),
        (0,),
        (0, 1),
        (0, 0),
        (4, 5),
        (2, 5, 2),
        (2, 2, 2, 2, 2, 2, 2, 2),
        5,
    ],
)
@pytest.mark.parametrize("usm_type", ["shared", "host", "device"])
def test_allocate_usm_ndarray(shape, usm_type):
    q = get_queue_or_skip()
    X = dpt.usm_ndarray(
        shape, dtype="i8", buffer=usm_type, buffer_ctor_kwargs={"queue": q}
    )
    Xnp = np.ndarray(shape, dtype="i8")
    assert X.usm_type == usm_type
    assert X.sycl_context == q.sycl_context
    assert X.sycl_device == q.sycl_device
    assert X.size == Xnp.size
    assert X.shape == Xnp.shape
    assert X.shape == X.__sycl_usm_array_interface__["shape"]


def test_usm_ndarray_flags():
    assert dpt.usm_ndarray((5,), dtype="i4").flags.fc
    assert dpt.usm_ndarray((5, 2), dtype="i4").flags.c_contiguous
    assert dpt.usm_ndarray((5, 2), dtype="i4", order="F").flags.f_contiguous
    assert dpt.usm_ndarray((5, 1, 2), dtype="i4", order="F").flags.f_contiguous
    assert dpt.usm_ndarray(
        (5, 1, 2), dtype="i4", strides=(2, 0, 1)
    ).flags.c_contiguous
    assert dpt.usm_ndarray(
        (5, 1, 2), dtype="i4", strides=(1, 0, 5)
    ).flags.f_contiguous
    assert dpt.usm_ndarray((5, 1, 1), dtype="i4", strides=(1, 0, 1)).flags.fc


@pytest.mark.parametrize(
    "dtype",
    [
        "u1",
        "i1",
        "u2",
        "i2",
        "u4",
        "i4",
        "u8",
        "i8",
        "f2",
        "f4",
        "f8",
        "c8",
        "c16",
        b"float32",
        dpt.dtype("d"),
        np.half,
    ],
)
def test_dtypes(dtype):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)
    Xusm = dpt.usm_ndarray((1,), dtype=dtype)
    assert Xusm.itemsize == dpt.dtype(dtype).itemsize
    expected_fmt = (dpt.dtype(dtype).str)[1:]
    actual_fmt = Xusm.__sycl_usm_array_interface__["typestr"][1:]
    assert expected_fmt == actual_fmt


@pytest.mark.parametrize(
    "dtype",
    [
        "",
        ">f4",
        "invalid",
        123,
        np.dtype(">f4"),
        np.dtype([("a", ">f4"), ("b", "i4")]),
    ],
)
def test_dtypes_invalid(dtype):
    with pytest.raises((TypeError, ValueError)):
        dpt.usm_ndarray((1,), dtype=dtype)


@pytest.mark.parametrize("dt", ["f", "c8"])
def test_properties(dt):
    """
    Test that properties execute
    """
    X = dpt.usm_ndarray((3, 4, 5), dtype=dt)
    assert isinstance(X.sycl_queue, dpctl.SyclQueue)
    assert isinstance(X.sycl_device, dpctl.SyclDevice)
    assert isinstance(X.sycl_context, dpctl.SyclContext)
    assert isinstance(X.dtype, dpt.dtype)
    assert isinstance(X.__sycl_usm_array_interface__, dict)
    assert isinstance(X.mT, dpt.usm_ndarray)
    assert isinstance(X.imag, dpt.usm_ndarray)
    assert isinstance(X.real, dpt.usm_ndarray)
    assert isinstance(X.shape, tuple)
    assert isinstance(X.strides, tuple)
    assert X.usm_type in ("shared", "device", "host")
    assert isinstance(X.size, numbers.Integral)
    assert isinstance(X.nbytes, numbers.Integral)
    assert isinstance(X.ndim, numbers.Integral)
    assert isinstance(X._pointer, numbers.Integral)
    assert isinstance(X.device, Device)
    with pytest.raises(ValueError):
        # array-API mandates exception for .ndim != 2
        X.T
    Y = dpt.usm_ndarray((2, 3), dtype=dt)
    assert isinstance(Y.mT, dpt.usm_ndarray)
    V = dpt.usm_ndarray((3,), dtype=dt)
    with pytest.raises(ValueError):
        # array-API mandates exception for .ndim != 2
        V.mT


@pytest.mark.parametrize("func", [bool, float, int, complex])
@pytest.mark.parametrize("shape", [tuple(), (1,), (1, 1), (1, 1, 1)])
@pytest.mark.parametrize("dtype", ["|b1", "|u2", "|f4", "|i8"])
def test_copy_scalar_with_func(func, shape, dtype):
    X = dpt.usm_ndarray(shape, dtype=dtype)
    Y = np.arange(1, X.size + 1, dtype=dtype).reshape(shape)
    X.usm_data.copy_from_host(Y.reshape(-1).view("|u1"))
    assert func(X) == func(Y)


@pytest.mark.parametrize(
    "method", ["__bool__", "__float__", "__int__", "__complex__"]
)
@pytest.mark.parametrize("shape", [tuple(), (1,), (1, 1), (1, 1, 1)])
@pytest.mark.parametrize("dtype", ["|b1", "|u2", "|f4", "|i8"])
def test_copy_scalar_with_method(method, shape, dtype):
    X = dpt.usm_ndarray(shape, dtype=dtype)
    Y = np.arange(1, X.size + 1, dtype=dtype).reshape(shape)
    X.usm_data.copy_from_host(Y.reshape(-1).view("|u1"))
    assert getattr(X, method)() == getattr(Y, method)()


@pytest.mark.parametrize("func", [bool, float, int, complex])
@pytest.mark.parametrize("shape", [(2,), (1, 2), (3, 4, 5), (0,)])
def test_copy_scalar_invalid_shape(func, shape):
    X = dpt.usm_ndarray(shape, dtype="i8")
    with pytest.raises(ValueError):
        func(X)


def test_index_noninteger():
    import operator

    X = dpt.usm_ndarray(1, "f4")
    with pytest.raises(IndexError):
        operator.index(X)


@pytest.mark.parametrize(
    "ind",
    [
        tuple(),
        (None,),
        (
            None,
            Ellipsis,
            None,
        ),
        (2, 2, None, 3, 4),
        (Ellipsis,),
        (None, slice(0, None, 2), Ellipsis, slice(0, None, 3)),
        (None, slice(1, None, 2), Ellipsis, slice(1, None, 3)),
        (None, slice(None, -1, -2), Ellipsis, slice(2, None, 3)),
        (
            slice(None, None, -1),
            slice(None, None, -1),
            slice(0, None, 3),
            slice(1, None, 2),
        ),
    ],
)
def test_basic_slice(ind):
    X = dpt.usm_ndarray((2 * 3, 2 * 4, 3 * 5, 2 * 7), dtype="u1")
    Xnp = np.empty(X.shape, dtype=X.dtype)
    S = X[ind]
    Snp = Xnp[ind]
    assert S.shape == Snp.shape
    assert S.strides == Snp.strides
    assert S.dtype == X.dtype


def test_empty_slice():
    # see gh801
    X = dpt.empty((1, 0, 1), dtype="u1")
    Y = X[:, ::-1, :]
    assert Y.shape == X.shape
    Z = X[:, ::2, :]
    assert Z.shape == X.shape
    X = dpt.empty(0)
    Y = X[::-1]
    assert Y.shape == X.shape
    Z = X[::2]
    assert Z.shape == X.shape
    X = dpt.empty((0, 4), dtype="u1")
    assert X[:, 1].shape == (0,)
    assert X[:, 1:3].shape == (0, 2)


def test_slice_constructor_1d():
    Xh = np.arange(37, dtype="i4")
    default_device = dpctl.select_default_device()
    Xusm = dpt.from_numpy(Xh, device=default_device, usm_type="device")
    for ind in [
        slice(1, None, 2),
        slice(0, None, 3),
        slice(1, None, 3),
        slice(2, None, 3),
        slice(None, None, -1),
        slice(-2, 2, -2),
        slice(-1, 1, -2),
        slice(None, None, -13),
    ]:
        assert np.array_equal(
            dpt.asnumpy(Xusm[ind]), Xh[ind]
        ), "Failed for {}".format(ind)


def test_slice_constructor_3d():
    Xh = np.empty((37, 24, 35), dtype="i4")
    default_device = dpctl.select_default_device()
    Xusm = dpt.from_numpy(Xh, device=default_device, usm_type="device")
    for ind in [
        slice(1, None, 2),
        slice(0, None, 3),
        slice(1, None, 3),
        slice(2, None, 3),
        slice(None, None, -1),
        slice(-2, 2, -2),
        slice(-1, 1, -2),
        slice(None, None, -13),
        (slice(None, None, -2), Ellipsis, None, 15),
    ]:
        assert np.array_equal(
            dpt.to_numpy(Xusm[ind]), Xh[ind]
        ), "Failed for {}".format(ind)


@pytest.mark.parametrize("usm_type", ["device", "shared", "host"])
def test_slice_suai(usm_type):
    Xh = np.arange(0, 10, dtype="u1")
    default_device = dpctl.select_default_device()
    Xusm = dpt.from_numpy(Xh, device=default_device, usm_type=usm_type)
    for ind in [slice(2, 3, None), slice(5, 7, None), slice(3, 9, None)]:
        assert np.array_equal(
            dpm.as_usm_memory(Xusm[ind]).copy_to_host(), Xh[ind]
        ), "Failed for {}".format(ind)


def test_slicing_basic():
    Xusm = dpt.usm_ndarray((10, 5), dtype="c8")
    Xusm[None]
    Xusm[...]
    Xusm[8]
    Xusm[-3]
    with pytest.raises(IndexError):
        Xusm[..., ...]
    with pytest.raises(IndexError):
        Xusm[1, 1, :, 1]
    Xusm[:, -4]
    with pytest.raises(IndexError):
        Xusm[:, -128]
    with pytest.raises(TypeError):
        Xusm[{1, 2, 3, 4, 5, 6, 7}]
    X = dpt.usm_ndarray(10, "u1")
    X.usm_data.copy_from_host(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09")
    int(
        X[X[2]]
    )  # check that objects with __index__ method can be used as indices
    Xh = dpm.as_usm_memory(X[X[2] : X[5]]).copy_to_host()
    Xnp = np.arange(0, 10, dtype="u1")
    assert np.array_equal(Xh, Xnp[Xnp[2] : Xnp[5]])


def test_ctor_invalid_shape():
    with pytest.raises(TypeError):
        dpt.usm_ndarray(dict())


def test_ctor_invalid_order():
    with pytest.raises(ValueError):
        dpt.usm_ndarray((5, 5, 3), order="Z")


def test_ctor_buffer_kwarg():
    dpt.usm_ndarray(10, dtype="i8", buffer=b"device")
    with pytest.raises(ValueError):
        dpt.usm_ndarray(10, buffer="invalid_param")
    Xusm = dpt.usm_ndarray((10, 5), dtype="c8")
    X2 = dpt.usm_ndarray(Xusm.shape, buffer=Xusm, dtype=Xusm.dtype)
    assert np.array_equal(
        Xusm.usm_data.copy_to_host(), X2.usm_data.copy_to_host()
    )
    with pytest.raises(ValueError):
        dpt.usm_ndarray(10, dtype="i4", buffer=dict())


def test_usm_ndarray_props():
    Xusm = dpt.usm_ndarray((10, 5), dtype="c8", order="F")
    Xusm.ndim
    repr(Xusm)
    Xusm.flags
    Xusm.__sycl_usm_array_interface__
    Xusm.device
    Xusm.strides
    Xusm.real
    Xusm.imag
    try:
        dpctl.SyclQueue("cpu")
    except dpctl.SyclQueueCreationError:
        pytest.skip("Sycl device CPU was not detected")
    Xusm.to_device("cpu")


def test_datapi_device():
    X = dpt.usm_ndarray(1, dtype="i4")
    dev_t = type(X.device)
    with pytest.raises(TypeError):
        dev_t()
    dev_t.create_device(X.device)
    dev_t.create_device(X.sycl_queue)
    d1 = dev_t.create_device(X.sycl_device)
    d2 = dev_t.create_device(X.sycl_device.filter_string)
    d3 = dev_t.create_device(None)
    assert d1.sycl_queue == d2.sycl_queue
    assert d1.sycl_queue == d3.sycl_queue
    X.device.sycl_context
    X.device.sycl_queue
    X.device.sycl_device
    repr(X.device)
    X.device.print_device_info()


def _pyx_capi_fnptr_to_callable(
    X,
    pyx_capi_name,
    caps_name,
    fn_restype=ctypes.c_void_p,
    fn_argtypes=(ctypes.py_object,),
):
    import sys

    mod = sys.modules[X.__class__.__module__]
    cap = mod.__pyx_capi__.get(pyx_capi_name, None)
    if cap is None:
        raise ValueError(
            "__pyx_capi__ does not export {} capsule".format(pyx_capi_name)
        )
    # construct Python callable to invoke these functions
    cap_ptr_fn = ctypes.pythonapi.PyCapsule_GetPointer
    cap_ptr_fn.restype = ctypes.c_void_p
    cap_ptr_fn.argtypes = [ctypes.py_object, ctypes.c_char_p]
    fn_ptr = cap_ptr_fn(cap, caps_name)
    callable_maker_ptr = ctypes.PYFUNCTYPE(fn_restype, *fn_argtypes)
    return callable_maker_ptr(fn_ptr)


def test_pyx_capi_get_data():
    X = dpt.usm_ndarray(17, dtype="i8")[1::2]
    get_data_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetData",
        b"char *(struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_void_p,
        fn_argtypes=(ctypes.py_object,),
    )
    r1 = get_data_fn(X)
    sua_iface = X.__sycl_usm_array_interface__
    assert r1 == sua_iface["data"][0] + sua_iface.get("offset") * X.itemsize


def test_pyx_capi_get_shape():
    X = dpt.usm_ndarray(17, dtype="u4")[1::2]
    get_shape_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetShape",
        b"Py_ssize_t *(struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_void_p,
        fn_argtypes=(ctypes.py_object,),
    )
    c_longlong_p = ctypes.POINTER(ctypes.c_longlong)
    shape0 = ctypes.cast(get_shape_fn(X), c_longlong_p).contents.value
    assert shape0 == X.shape[0]


def test_pyx_capi_get_strides():
    X = dpt.usm_ndarray(17, dtype="f4")[1::2]
    get_strides_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetStrides",
        b"Py_ssize_t *(struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_void_p,
        fn_argtypes=(ctypes.py_object,),
    )
    c_longlong_p = ctypes.POINTER(ctypes.c_longlong)
    strides0_p = get_strides_fn(X)
    if strides0_p:
        strides0_p = ctypes.cast(strides0_p, c_longlong_p).contents
        strides0_p = strides0_p.value
    assert strides0_p == 0 or strides0_p == X.strides[0]


def test_pyx_capi_get_ndim():
    X = dpt.usm_ndarray(17, dtype="?")[1::2]
    get_ndim_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetNDim",
        b"int (struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_int,
        fn_argtypes=(ctypes.py_object,),
    )
    assert get_ndim_fn(X) == X.ndim


def test_pyx_capi_get_typenum():
    X = dpt.usm_ndarray(17, dtype="c8")[1::2]
    get_typenum_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetTypenum",
        b"int (struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_int,
        fn_argtypes=(ctypes.py_object,),
    )
    typenum = get_typenum_fn(X)
    assert type(typenum) is int
    assert typenum == X.dtype.num


def test_pyx_capi_get_elemsize():
    X = dpt.usm_ndarray(17, dtype="u8")[1::2]
    get_elemsize_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetElementSize",
        b"int (struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_int,
        fn_argtypes=(ctypes.py_object,),
    )
    itemsize = get_elemsize_fn(X)
    assert type(itemsize) is int
    assert itemsize == X.itemsize


def test_pyx_capi_get_flags():
    X = dpt.usm_ndarray(17, dtype="i8")[1::2]
    get_flags_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetFlags",
        b"int (struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_int,
        fn_argtypes=(ctypes.py_object,),
    )
    flags = get_flags_fn(X)
    assert type(flags) is int and X.flags == flags


def test_pyx_capi_get_offset():
    X = dpt.usm_ndarray(17, dtype="u2")[1::2]
    get_offset_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetOffset",
        b"Py_ssize_t (struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_longlong,
        fn_argtypes=(ctypes.py_object,),
    )
    offset = get_offset_fn(X)
    assert type(offset) is int
    assert offset == X.__sycl_usm_array_interface__["offset"]


def test_pyx_capi_get_queue_ref():
    X = dpt.usm_ndarray(17, dtype="i2")[1::2]
    get_queue_ref_fn = _pyx_capi_fnptr_to_callable(
        X,
        "UsmNDArray_GetQueueRef",
        b"DPCTLSyclQueueRef (struct PyUSMArrayObject *)",
        fn_restype=ctypes.c_void_p,
        fn_argtypes=(ctypes.py_object,),
    )
    queue_ref = get_queue_ref_fn(X)  # address of a copy, should be unequal
    assert queue_ref != X.sycl_queue.addressof_ref()


def test_pyx_capi_make_from_memory():
    q = get_queue_or_skip()
    n0, n1 = 4, 6
    c_tuple = (ctypes.c_ssize_t * 2)(n0, n1)
    mem = dpm.MemoryUSMShared(n0 * n1 * 4, queue=q)
    typenum = dpt.dtype("single").num
    any_usm_ndarray = dpt.empty(tuple(), dtype="i4", sycl_queue=q)
    make_from_memory_fn = _pyx_capi_fnptr_to_callable(
        any_usm_ndarray,
        "UsmNDArray_MakeSimpleFromMemory",
        b"PyObject *(int, Py_ssize_t const *, int, "
        b"struct Py_MemoryObject *, Py_ssize_t, char)",
        fn_restype=ctypes.py_object,
        fn_argtypes=(
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ssize_t),
            ctypes.c_int,
            ctypes.py_object,
            ctypes.c_ssize_t,
            ctypes.c_char,
        ),
    )
    r = make_from_memory_fn(
        ctypes.c_int(2),
        c_tuple,
        ctypes.c_int(typenum),
        mem,
        ctypes.c_ssize_t(0),
        ctypes.c_char(b"C"),
    )
    assert isinstance(r, dpt.usm_ndarray)
    assert r.ndim == 2
    assert r.shape == (n0, n1)
    assert r._pointer == mem._pointer
    assert r.usm_type == "shared"
    assert r.sycl_queue == q
    assert r.flags["C"]
    r2 = make_from_memory_fn(
        ctypes.c_int(2),
        c_tuple,
        ctypes.c_int(typenum),
        mem,
        ctypes.c_ssize_t(0),
        ctypes.c_char(b"F"),
    )
    ptr = mem._pointer
    del mem
    del r
    assert isinstance(r2, dpt.usm_ndarray)
    assert r2._pointer == ptr
    assert r2.usm_type == "shared"
    assert r2.sycl_queue == q
    assert r2.flags["F"]


def test_pyx_capi_set_writable_flag():
    q = get_queue_or_skip()
    usm_ndarray = dpt.empty((4, 5), dtype="i4", sycl_queue=q)
    assert isinstance(usm_ndarray, dpt.usm_ndarray)
    assert usm_ndarray.flags["WRITABLE"] is True
    set_writable = _pyx_capi_fnptr_to_callable(
        usm_ndarray,
        "UsmNDArray_SetWritableFlag",
        b"void (struct PyUSMArrayObject *, int)",
        fn_restype=None,
        fn_argtypes=(ctypes.py_object, ctypes.c_int),
    )
    set_writable(usm_ndarray, ctypes.c_int(0))
    assert isinstance(usm_ndarray, dpt.usm_ndarray)
    assert usm_ndarray.flags["WRITABLE"] is False
    set_writable(usm_ndarray, ctypes.c_int(1))
    assert isinstance(usm_ndarray, dpt.usm_ndarray)
    assert usm_ndarray.flags["WRITABLE"] is True


def test_pyx_capi_make_from_ptr():
    q = get_queue_or_skip()
    usm_ndarray = dpt.empty(tuple(), dtype="i4", sycl_queue=q)
    make_from_ptr = _pyx_capi_fnptr_to_callable(
        usm_ndarray,
        "UsmNDArray_MakeSimpleFromPtr",
        b"PyObject *(size_t, int, DPCTLSyclUSMRef, "
        b"DPCTLSyclQueueRef, PyObject *)",
        fn_restype=ctypes.py_object,
        fn_argtypes=(
            ctypes.c_size_t,
            ctypes.c_int,
            ctypes.c_void_p,
            ctypes.c_void_p,
            ctypes.py_object,
        ),
    )
    nelems = 10
    dt = dpt.int64
    mem = dpm.MemoryUSMDevice(nelems * dt.itemsize, queue=q)
    arr = make_from_ptr(
        ctypes.c_size_t(nelems),
        dt.num,
        mem._pointer,
        mem.sycl_queue.addressof_ref(),
        mem,
    )
    assert isinstance(arr, dpt.usm_ndarray)
    assert arr.shape == (nelems,)
    assert arr.dtype == dt
    assert arr.sycl_queue == q
    assert arr._pointer == mem._pointer
    del mem
    assert isinstance(arr.__repr__(), str)


def test_pyx_capi_make_general():
    q = get_queue_or_skip()
    usm_ndarray = dpt.empty(tuple(), dtype="i4", sycl_queue=q)
    make_from_ptr = _pyx_capi_fnptr_to_callable(
        usm_ndarray,
        "UsmNDArray_MakeFromPtr",
        b"PyObject *(int, Py_ssize_t const *, int, Py_ssize_t const *, "
        b"DPCTLSyclUSMRef, DPCTLSyclQueueRef, Py_ssize_t, PyObject *)",
        fn_restype=ctypes.py_object,
        fn_argtypes=(
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ssize_t),
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_ssize_t),
            ctypes.c_void_p,
            ctypes.c_void_p,
            ctypes.c_ssize_t,
            ctypes.py_object,
        ),
    )
    # Create array to view into diagonal of a matrix
    n = 5
    mat = dpt.reshape(
        dpt.arange(n * n, dtype="i4", sycl_queue=q),
        (
            n,
            n,
        ),
    )
    c_shape = (ctypes.c_ssize_t * 1)(
        n,
    )
    c_strides = (ctypes.c_ssize_t * 1)(
        n + 1,
    )
    diag = make_from_ptr(
        ctypes.c_int(1),
        c_shape,
        ctypes.c_int(mat.dtype.num),
        c_strides,
        mat._pointer,
        mat.sycl_queue.addressof_ref(),
        ctypes.c_ssize_t(0),
        mat,
    )
    assert isinstance(diag, dpt.usm_ndarray)
    assert diag.shape == (n,)
    assert diag.strides == (n + 1,)
    assert diag.dtype == mat.dtype
    assert diag.sycl_queue == q
    assert diag._pointer == mat._pointer
    del mat
    assert isinstance(diag.__repr__(), str)
    # create 0d scalar
    mat = dpt.reshape(
        dpt.arange(n * n, dtype="i4", sycl_queue=q),
        (
            n,
            n,
        ),
    )
    sc = make_from_ptr(
        ctypes.c_int(0),
        None,  # NULL pointer
        ctypes.c_int(mat.dtype.num),
        None,  # NULL pointer
        mat._pointer,
        mat.sycl_queue.addressof_ref(),
        ctypes.c_ssize_t(0),
        mat,
    )
    assert isinstance(sc, dpt.usm_ndarray)
    assert sc.shape == tuple()
    assert sc.dtype == mat.dtype
    assert sc.sycl_queue == q
    assert sc._pointer == mat._pointer
    c_shape = (ctypes.c_ssize_t * 2)(0, n)
    c_strides = (ctypes.c_ssize_t * 2)(0, 1)
    zd_arr = make_from_ptr(
        ctypes.c_int(2),
        c_shape,
        ctypes.c_int(mat.dtype.num),
        c_strides,
        mat._pointer,
        mat.sycl_queue.addressof_ref(),
        ctypes.c_ssize_t(0),
        mat,
    )
    assert isinstance(zd_arr, dpt.usm_ndarray)
    assert zd_arr.shape == (
        0,
        n,
    )
    assert zd_arr.strides == (
        0,
        1,
    )
    assert zd_arr.dtype == mat.dtype
    assert zd_arr.sycl_queue == q
    assert zd_arr._pointer == mat._pointer


def _pyx_capi_int(X, pyx_capi_name, caps_name=b"int", val_restype=ctypes.c_int):
    import sys

    mod = sys.modules[X.__class__.__module__]
    cap = mod.__pyx_capi__.get(pyx_capi_name, None)
    if cap is None:
        raise ValueError(
            "__pyx_capi__ does not export {} capsule".format(pyx_capi_name)
        )
    # construct Python callable to invoke these functions
    cap_ptr_fn = ctypes.pythonapi.PyCapsule_GetPointer
    cap_ptr_fn.restype = ctypes.c_void_p
    cap_ptr_fn.argtypes = [ctypes.py_object, ctypes.c_char_p]
    cap_ptr = cap_ptr_fn(cap, caps_name)
    val_ptr = ctypes.cast(cap_ptr, ctypes.POINTER(val_restype))
    return val_ptr.contents.value


def test_pyx_capi_check_constants():
    X = dpt.usm_ndarray(17, dtype="i1")[1::2]
    cc_flag = _pyx_capi_int(X, "USM_ARRAY_C_CONTIGUOUS")
    assert cc_flag > 0 and 0 == (cc_flag & (cc_flag - 1))
    fc_flag = _pyx_capi_int(X, "USM_ARRAY_F_CONTIGUOUS")
    assert fc_flag > 0 and 0 == (fc_flag & (fc_flag - 1))
    w_flag = _pyx_capi_int(X, "USM_ARRAY_WRITABLE")
    assert w_flag > 0 and 0 == (w_flag & (w_flag - 1))

    bool_typenum = _pyx_capi_int(X, "UAR_BOOL")
    assert bool_typenum == dpt.dtype("bool_").num

    byte_typenum = _pyx_capi_int(X, "UAR_BYTE")
    assert byte_typenum == dpt.dtype(np.byte).num
    ubyte_typenum = _pyx_capi_int(X, "UAR_UBYTE")
    assert ubyte_typenum == dpt.dtype(np.ubyte).num

    short_typenum = _pyx_capi_int(X, "UAR_SHORT")
    assert short_typenum == dpt.dtype(np.short).num
    ushort_typenum = _pyx_capi_int(X, "UAR_USHORT")
    assert ushort_typenum == dpt.dtype(np.ushort).num

    int_typenum = _pyx_capi_int(X, "UAR_INT")
    assert int_typenum == dpt.dtype(np.intc).num
    uint_typenum = _pyx_capi_int(X, "UAR_UINT")
    assert uint_typenum == dpt.dtype(np.uintc).num

    long_typenum = _pyx_capi_int(X, "UAR_LONG")
    assert long_typenum == dpt.dtype(np.int_).num
    ulong_typenum = _pyx_capi_int(X, "UAR_ULONG")
    assert ulong_typenum == dpt.dtype(np.uint).num

    longlong_typenum = _pyx_capi_int(X, "UAR_LONGLONG")
    assert longlong_typenum == dpt.dtype(np.longlong).num
    ulonglong_typenum = _pyx_capi_int(X, "UAR_ULONGLONG")
    assert ulonglong_typenum == dpt.dtype(np.ulonglong).num

    half_typenum = _pyx_capi_int(X, "UAR_HALF")
    assert half_typenum == dpt.dtype(np.half).num
    float_typenum = _pyx_capi_int(X, "UAR_FLOAT")
    assert float_typenum == dpt.dtype(np.single).num
    double_typenum = _pyx_capi_int(X, "UAR_DOUBLE")
    assert double_typenum == dpt.dtype(np.double).num

    cfloat_typenum = _pyx_capi_int(X, "UAR_CFLOAT")
    assert cfloat_typenum == dpt.dtype(np.csingle).num
    cdouble_typenum = _pyx_capi_int(X, "UAR_CDOUBLE")
    assert cdouble_typenum == dpt.dtype(np.cdouble).num


_all_dtypes = [
    "b1",
    "i1",
    "u1",
    "i2",
    "u2",
    "i4",
    "u4",
    "i8",
    "u8",
    "f2",
    "f4",
    "f8",
    "c8",
    "c16",
]


@pytest.mark.parametrize(
    "shape", [tuple(), (1,), (5,), (2, 3), (2, 3, 4), (2, 2, 2, 2, 2)]
)
@pytest.mark.parametrize(
    "dtype",
    _all_dtypes,
)
@pytest.mark.parametrize("usm_type", ["device", "shared", "host"])
def test_tofrom_numpy(shape, dtype, usm_type):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)
    Xnp = np.zeros(shape, dtype=dtype)
    Xusm = dpt.from_numpy(Xnp, usm_type=usm_type, sycl_queue=q)
    Ynp = np.ones(shape, dtype=dtype)
    ind = (slice(None, None, None),) * Ynp.ndim
    Xusm[ind] = Ynp
    assert np.array_equal(dpt.to_numpy(Xusm), Ynp)


@pytest.mark.parametrize(
    "dtype",
    _all_dtypes,
)
@pytest.mark.parametrize("src_usm_type", ["device", "shared", "host"])
@pytest.mark.parametrize("dst_usm_type", ["device", "shared", "host"])
def test_setitem_same_dtype(dtype, src_usm_type, dst_usm_type):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)
    Xnp = (
        np.random.randint(-10, 10, size=2 * 3 * 4)
        .astype(dtype)
        .reshape((2, 4, 3))
    )
    Znp = np.zeros(
        (
            2,
            4,
            3,
        ),
        dtype=dtype,
    )
    Zusm_0d = dpt.from_numpy(Znp[0, 0, 0], usm_type=dst_usm_type)
    ind = (-1, -1, -1)
    Xusm_0d = dpt.from_numpy(Xnp[ind], usm_type=src_usm_type)
    Zusm_0d[Ellipsis] = Xusm_0d
    assert np.array_equal(dpt.to_numpy(Zusm_0d), Xnp[ind])
    Zusm_1d = dpt.from_numpy(Znp[0, 1:3, 0], usm_type=dst_usm_type)
    ind = (-1, slice(0, 2, None), -1)
    Xusm_1d = dpt.from_numpy(Xnp[ind], usm_type=src_usm_type)
    Zusm_1d[Ellipsis] = Xusm_1d
    assert np.array_equal(dpt.to_numpy(Zusm_1d), Xnp[ind])
    Zusm_2d = dpt.from_numpy(Znp[:, 1:3, 0], usm_type=dst_usm_type)[::-1]
    Xusm_2d = dpt.from_numpy(Xnp[:, 1:4, -1], usm_type=src_usm_type)
    Zusm_2d[:] = Xusm_2d[:, 0:2]
    assert np.array_equal(dpt.to_numpy(Zusm_2d), Xnp[:, 1:3, -1])
    Zusm_3d = dpt.from_numpy(Znp, usm_type=dst_usm_type)
    Xusm_3d = dpt.from_numpy(Xnp, usm_type=src_usm_type)
    Zusm_3d[:] = Xusm_3d
    assert np.array_equal(dpt.to_numpy(Zusm_3d), Xnp)
    Zusm_3d[::-1] = Xusm_3d[::-1]
    assert np.array_equal(dpt.to_numpy(Zusm_3d), Xnp)
    Zusm_3d[:] = Xusm_3d[0]
    R1 = dpt.to_numpy(Zusm_3d)
    R2 = np.broadcast_to(Xnp[0], R1.shape)
    assert R1.shape == R2.shape
    assert np.allclose(R1, R2)
    Zusm_empty = Zusm_1d[0:0]
    Zusm_empty[Ellipsis] = Zusm_3d[0, 0, 0:0]


@pytest.mark.parametrize(
    "dtype",
    _all_dtypes,
)
@pytest.mark.parametrize("usm_type", ["device", "shared", "host"])
def test_setitem_scalar(dtype, usm_type):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)
    X = dpt.usm_ndarray((6, 6), dtype=dtype, buffer=usm_type)
    for i in range(X.size):
        X[np.unravel_index(i, X.shape)] = np.asarray(i, dtype=dtype)
    assert np.array_equal(
        dpt.to_numpy(X), np.arange(X.size).astype(dtype).reshape(X.shape)
    )
    Y = dpt.usm_ndarray((2, 3), dtype=dtype, buffer=usm_type)
    for i in range(Y.size):
        Y[np.unravel_index(i, Y.shape)] = i
    assert np.array_equal(
        dpt.to_numpy(Y), np.arange(Y.size).astype(dtype).reshape(Y.shape)
    )


def test_setitem_errors():
    q = get_queue_or_skip()
    X = dpt.empty((4,), dtype="u1", sycl_queue=q)
    Y = dpt.empty((4, 2), dtype="u1", sycl_queue=q)
    with pytest.raises(ValueError):
        X[:] = Y
    with pytest.raises(ValueError):
        X[:] = Y[:, 0:1]
    X[:] = Y[None, :, 0]


@pytest.mark.parametrize("src_dt,dst_dt", [("i4", "i8"), ("f4", "f8")])
def test_setitem_different_dtypes(src_dt, dst_dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dst_dt, q)
    X = dpt.from_numpy(np.ones(10, src_dt), sycl_queue=q)
    Y = dpt.from_numpy(np.zeros(10, src_dt), sycl_queue=q)
    Z = dpt.empty((20,), dtype=dst_dt, sycl_queue=q)
    Z[::2] = X
    Z[1::2] = Y
    assert np.allclose(dpt.asnumpy(Z), np.tile(np.array([1, 0], Z.dtype), 10))


def test_setitem_wingaps():
    q = get_queue_or_skip()
    if dpt.dtype("intc").itemsize == dpt.dtype("int32").itemsize:
        dpt_dst = dpt.empty(4, dtype="int32", sycl_queue=q)
        np_src = np.arange(4, dtype="intc")
        dpt_dst[:] = np_src  # should not raise exceptions
        assert np.array_equal(dpt.asnumpy(dpt_dst), np_src)
    if dpt.dtype("long").itemsize == dpt.dtype("longlong").itemsize:
        dpt_dst = dpt.empty(4, dtype="longlong", sycl_queue=q)
        np_src = np.arange(4, dtype="long")
        dpt_dst[:] = np_src  # should not raise exceptions
        assert np.array_equal(dpt.asnumpy(dpt_dst), np_src)


def test_shape_setter():
    def cc_strides(sh):
        return np.empty(sh, dtype="u1").strides

    def relaxed_strides_equal(st1, st2, sh):
        eq_ = True
        for s1, s2, d in zip(st1, st2, sh):
            eq_ = eq_ and ((d == 1) or (s1 == s2))
        return eq_

    sh_s = (2 * 3 * 4 * 5,)
    sh_f = (
        2,
        3,
        4,
        5,
    )
    X = dpt.usm_ndarray(sh_s, dtype="i8")
    X.shape = sh_f
    assert X.shape == sh_f
    assert relaxed_strides_equal(X.strides, cc_strides(sh_f), sh_f)
    assert X.flags.c_contiguous, "reshaped array expected to be C-contiguous"

    sh_s = (
        2,
        12,
        5,
    )
    sh_f = (
        2,
        3,
        4,
        5,
    )
    X = dpt.usm_ndarray(sh_s, dtype="u4", order="C")
    X.shape = sh_f
    assert X.shape == sh_f
    assert relaxed_strides_equal(X.strides, cc_strides(sh_f), sh_f)

    sh_s = (2, 3, 4, 5)
    sh_f = (4, 3, 2, 5)
    X = dpt.usm_ndarray(sh_s, dtype="f4")
    X.shape = sh_f
    assert relaxed_strides_equal(X.strides, cc_strides(sh_f), sh_f)

    sh_s = (2, 3, 4, 5)
    sh_f = (4, 3, 1, 2, 5)
    X = dpt.usm_ndarray(sh_s, dtype="?")
    X.shape = sh_f
    assert relaxed_strides_equal(X.strides, cc_strides(sh_f), sh_f)

    X = dpt.usm_ndarray(sh_s, dtype="u4")
    with pytest.raises(TypeError):
        X.shape = "abcbe"
    X = dpt.usm_ndarray((4, 4), dtype="u1")[::2, ::2]
    with pytest.raises(AttributeError):
        X.shape = (4,)
    X = dpt.usm_ndarray((0,), dtype="i4")
    X.shape = (0,)
    X.shape = (
        2,
        0,
    )
    X.shape = (
        0,
        2,
    )
    X.shape = (
        1,
        0,
        1,
    )


def test_len():
    X = dpt.usm_ndarray(1, "i4")
    assert len(X) == 1
    X = dpt.usm_ndarray((2, 1), "i4")
    assert len(X) == 2
    X = dpt.usm_ndarray(tuple(), "i4")
    with pytest.raises(TypeError):
        len(X)


def test_array_namespace():
    X = dpt.usm_ndarray(1, "i4")
    X.__array_namespace__()
    X._set_namespace(dpt)
    assert X.__array_namespace__() is dpt


def test_dlpack():
    X = dpt.usm_ndarray(1, "i4")
    X.__dlpack_device__()
    X.__dlpack__(stream=None)


def test_to_device():
    X = dpt.usm_ndarray(1, "f4")
    for dev in dpctl.get_devices():
        if dev.default_selector_score > 0:
            Y = X.to_device(dev)
            assert Y.sycl_device == dev


def test_to_device_migration():
    q1 = get_queue_or_skip()  # two distinct copies of default-constructed queue
    q2 = get_queue_or_skip()
    X1 = dpt.empty((5,), "i8", sycl_queue=q1)  # X1 is associated with q1
    X2 = X1.to_device(q2)  # X2 is reassociated with q2
    assert X1.sycl_queue == q1
    assert X2.sycl_queue == q2
    assert X1.usm_data._pointer == X2.usm_data._pointer


def test_astype():
    X = dpt.empty((5, 5), dtype="i4")
    X[:] = np.full((5, 5), 7, dtype="i4")
    Y = dpt.astype(X, "c8", order="C")
    assert np.allclose(dpt.to_numpy(Y), np.full((5, 5), 7, dtype="c8"))
    if Y.sycl_device.has_aspect_fp16:
        Y = dpt.astype(X[::2, ::-1], "f2", order="K")
        assert np.allclose(dpt.to_numpy(Y), np.full(Y.shape, 7, dtype="f2"))
    Y = dpt.astype(X[::2, ::-1], "f4", order="K")
    assert np.allclose(dpt.to_numpy(Y), np.full(Y.shape, 7, dtype="f4"))
    Y = dpt.astype(X[::2, ::-1], "i4", order="K", copy=False)
    assert Y.usm_data is X.usm_data


def test_astype_invalid_order():
    X = dpt.usm_ndarray(5, "i4")
    with pytest.raises(ValueError):
        dpt.astype(X, "i4", order="WRONG")


def test_copy():
    X = dpt.usm_ndarray((5, 5), "i4")[2:4, 1:4]
    X[:] = 42
    Yc = dpt.copy(X, order="C")
    Yf = dpt.copy(X, order="F")
    Ya = dpt.copy(X, order="A")
    Yk = dpt.copy(X, order="K")
    assert Yc.usm_data is not X.usm_data
    assert Yf.usm_data is not X.usm_data
    assert Ya.usm_data is not X.usm_data
    assert Yk.usm_data is not X.usm_data
    assert Yc.strides == (3, 1)
    assert Yf.strides == (1, 2)
    assert Ya.strides == (3, 1)
    assert Yk.strides == (3, 1)
    ref = np.full(X.shape, 42, dtype=X.dtype)
    assert np.array_equal(dpt.asnumpy(Yc), ref)
    assert np.array_equal(dpt.asnumpy(Yf), ref)
    assert np.array_equal(dpt.asnumpy(Ya), ref)
    assert np.array_equal(dpt.asnumpy(Yk), ref)


def test_ctor_invalid():
    m = dpm.MemoryUSMShared(12)
    with pytest.raises(ValueError):
        dpt.usm_ndarray((4,), dtype="i4", buffer=m)
    m = dpm.MemoryUSMShared(64)
    with pytest.raises(ValueError):
        dpt.usm_ndarray((4,), dtype="u1", buffer=m, strides={"not": "valid"})


def test_reshape():
    X = dpt.usm_ndarray((5, 5), "i4")
    # can be done as views
    Y = dpt.reshape(X, (25,))
    assert Y.shape == (25,)
    Z = X[::2, ::2]
    # requires a copy
    W = dpt.reshape(Z, (Z.size,), order="F")
    assert W.shape == (Z.size,)
    with pytest.raises(TypeError):
        dpt.reshape("invalid")
    with pytest.raises(ValueError):
        dpt.reshape(Z, (2, 2, 2, 2, 2))
    with pytest.raises(ValueError):
        dpt.reshape(Z, Z.shape, order="invalid")
    W = dpt.reshape(Z, (-1,), order="C")
    assert W.shape == (Z.size,)

    X = dpt.usm_ndarray((1,), dtype="i8")
    Y = dpt.reshape(X, X.shape)
    assert Y.flags == X.flags

    A = dpt.usm_ndarray((0,), "i4")
    A1 = dpt.reshape(A, (0,))
    assert A1.shape == (0,)
    requested_shape = (
        2,
        0,
    )
    A2 = dpt.reshape(A, requested_shape)
    assert A2.shape == requested_shape
    requested_shape = (
        0,
        2,
    )
    A3 = dpt.reshape(A, requested_shape)
    assert A3.shape == requested_shape
    requested_shape = (
        1,
        0,
        2,
    )
    A4 = dpt.reshape(A, requested_shape)
    assert A4.shape == requested_shape


def test_reshape_copy_kwrd():
    X = dpt.usm_ndarray((2, 3), "i4")
    new_shape = (6,)
    Z = dpt.reshape(X, new_shape, copy=True)
    assert Z.shape == new_shape
    assert Z.usm_data is not X.usm_data
    X = dpt.usm_ndarray((3, 3), "i4")[::2, ::2]
    new_shape = (4,)
    with pytest.raises(ValueError):
        Z = dpt.reshape(X, new_shape, copy=False)
    with pytest.raises(ValueError):
        invalid = Ellipsis
        Z = dpt.reshape(X, new_shape, copy=invalid)


def test_transpose():
    n, m = 2, 3
    X = dpt.usm_ndarray((n, m), "f4")
    Xnp = np.arange(n * m, dtype="f4").reshape((n, m))
    X[:] = Xnp
    assert np.array_equal(dpt.to_numpy(X.T), Xnp.T)
    assert np.array_equal(dpt.to_numpy(X[1:].T), Xnp[1:].T)


def test_real_imag_views():
    n, m = 2, 3
    X = dpt.usm_ndarray((n, m), "c8")
    Xnp_r = np.arange(n * m, dtype="f4").reshape((n, m))
    Xnp_i = np.arange(n * m, 2 * n * m, dtype="f4").reshape((n, m))
    Xnp = Xnp_r + 1j * Xnp_i
    X[:] = Xnp
    assert np.array_equal(dpt.to_numpy(X.real), Xnp.real)
    assert np.array_equal(dpt.to_numpy(X.imag), Xnp.imag)
    assert np.array_equal(dpt.to_numpy(X[1:].real), Xnp[1:].real)
    assert np.array_equal(dpt.to_numpy(X[1:].imag), Xnp[1:].imag)


@pytest.mark.parametrize(
    "dtype",
    _all_dtypes,
)
def test_zeros(dtype):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)
    X = dpt.zeros(10, dtype=dtype, sycl_queue=q)
    assert np.array_equal(dpt.asnumpy(X), np.zeros(10, dtype=dtype))


@pytest.mark.parametrize(
    "dtype",
    _all_dtypes,
)
def test_ones(dtype):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)
    X = dpt.ones(10, dtype=dtype, sycl_queue=q)
    assert np.array_equal(dpt.asnumpy(X), np.ones(10, dtype=dtype))


@pytest.mark.parametrize(
    "dtype",
    _all_dtypes,
)
def test_full(dtype):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)
    X = dpt.full(10, 4, dtype=dtype, sycl_queue=q)
    assert np.array_equal(dpt.asnumpy(X), np.full(10, 4, dtype=dtype))


def test_full_dtype_inference():
    assert np.issubdtype(dpt.full(10, 4).dtype, np.integer)
    assert dpt.full(10, True).dtype is dpt.dtype(np.bool_)
    assert np.issubdtype(dpt.full(10, 12.3).dtype, np.floating)
    cdt = dpt.full(10, 0.3 - 2j).dtype
    assert np.issubdtype(cdt, np.complexfloating)

    assert np.issubdtype(dpt.full(10, 12.3, dtype=int).dtype, np.integer)
    assert np.issubdtype(dpt.full(10, 0.3 - 2j, dtype=int).dtype, np.integer)
    rdt = np.finfo(cdt).dtype
    assert np.issubdtype(dpt.full(10, 0.3 - 2j, dtype=rdt).dtype, np.floating)


def test_full_fill_array():
    q = get_queue_or_skip()

    Xnp = np.array([1, 2, 3], dtype="i4")
    X = dpt.asarray(Xnp, sycl_queue=q)

    shape = (3, 3)
    Y = dpt.full(shape, X)
    Ynp = np.full(shape, Xnp)

    assert Y.dtype == Ynp.dtype
    assert Y.usm_type == "device"
    assert np.array_equal(dpt.asnumpy(Y), Ynp)


def test_full_compute_follows_data():
    q1 = get_queue_or_skip()
    q2 = get_queue_or_skip()

    X = dpt.arange(10, dtype="i4", sycl_queue=q1, usm_type="shared")
    Y = dpt.full(10, X[3])

    assert Y.dtype == X.dtype
    assert Y.usm_type == X.usm_type
    assert dpctl.utils.get_execution_queue((Y.sycl_queue, X.sycl_queue))
    assert np.array_equal(dpt.asnumpy(Y), np.full(10, 3, dtype="i4"))

    Y = dpt.full(10, X[3], dtype="f4", sycl_queue=q2, usm_type="host")

    assert Y.dtype == dpt.dtype("f4")
    assert Y.usm_type == "host"
    assert dpctl.utils.get_execution_queue((Y.sycl_queue, q2))
    assert np.array_equal(dpt.asnumpy(Y), np.full(10, 3, dtype="f4"))


@pytest.mark.parametrize("order1", ["F", "C"])
@pytest.mark.parametrize("order2", ["F", "C"])
def test_full_order(order1, order2):
    q = get_queue_or_skip()
    Xnp = np.array([1, 2, 3], order=order1)
    Ynp = np.full((3, 3), Xnp, order=order2)
    Y = dpt.full((3, 3), Xnp, order=order2, sycl_queue=q)
    assert Y.flags.c_contiguous == Ynp.flags.c_contiguous
    assert Y.flags.f_contiguous == Ynp.flags.f_contiguous
    assert np.array_equal(dpt.asnumpy(Y), Ynp)


def test_full_strides():
    q = get_queue_or_skip()
    X = dpt.full((3, 3), dpt.arange(3, dtype="i4"), sycl_queue=q)
    Xnp = np.full((3, 3), np.arange(3, dtype="i4"))
    assert X.strides == tuple(el // Xnp.itemsize for el in Xnp.strides)
    assert np.array_equal(dpt.asnumpy(X), Xnp)

    X = dpt.full((3, 3), dpt.arange(6, dtype="i4")[::2], sycl_queue=q)
    Xnp = np.full((3, 3), np.arange(6, dtype="i4")[::2])
    assert X.strides == tuple(el // Xnp.itemsize for el in Xnp.strides)
    assert np.array_equal(dpt.asnumpy(X), Xnp)


@pytest.mark.parametrize(
    "dt",
    _all_dtypes[1:],
)
def test_arange(dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dt, q)
    X = dpt.arange(0, 123, dtype=dt, sycl_queue=q)
    dt = dpt.dtype(dt)
    if np.issubdtype(dt, np.integer):
        assert int(X[47]) == 47
    elif np.issubdtype(dt, np.floating):
        assert float(X[47]) == 47.0
    elif np.issubdtype(dt, np.complexfloating):
        assert complex(X[47]) == 47.0 + 0.0j

    # choose size larger than maximal value that u1/u2 can accomodate
    sz = int(dpt.iinfo(dpt.int8).max)
    X1 = dpt.arange(sz + 1, dtype=dt, sycl_queue=q)
    assert X1.shape == (sz + 1,)

    X2 = dpt.arange(sz, 0, -1, dtype=dt, sycl_queue=q)
    assert X2.shape == (sz,)


def test_arange_fp():
    q = get_queue_or_skip()

    assert dpt.arange(7, 0, -2, dtype="f4", device=q).shape == (4,)
    assert dpt.arange(0, 1, 0.25, dtype="f4", device=q).shape == (4,)

    has_fp64 = q.sycl_device.has_aspect_fp64
    if has_fp64:
        assert dpt.arange(7, 0, -2, dtype="f8", device=q).shape == (4,)
    assert dpt.arange(0, 1, 0.25, dtype="f4", device=q).shape == (4,)

    x = dpt.arange(9.7, stop=10, sycl_queue=q)
    assert x.shape == (1,)
    assert x.dtype == dpt.float64 if has_fp64 else dpt.float32


def test_arange_step_None():
    q = get_queue_or_skip()

    x = dpt.arange(0, stop=10, step=None, dtype="int32", sycl_queue=q)
    assert x.shape == (10,)


def test_arange_bool():
    q = get_queue_or_skip()

    x = dpt.arange(0, stop=2, dtype="bool", sycl_queue=q)
    assert x.shape == (2,)
    assert x.dtype == dpt.bool


def test_arange_mixed_types():
    q = get_queue_or_skip()

    x = dpt.arange(-2.5, stop=200, step=100, dtype="int32", sycl_queue=q)
    assert x.shape[0] == 3
    assert int(x[1]) == 99 + int(x[0])

    x = dpt.arange(+2.5, stop=200, step=100, dtype="int32", device=x.device)
    assert x.shape[0] == 2
    assert int(x[1]) == 100 + int(x[0])

    _stop = np.float32(504)
    x = dpt.arange(0, stop=_stop, step=100, dtype="f4", device=x.device)
    assert x.shape == (6,)

    # ensure length is determined using uncast parameters
    x = dpt.arange(-5, stop=10**2, step=2.7, dtype="int64", device=x.device)
    assert x.shape == (39,)


@pytest.mark.parametrize(
    "dt",
    _all_dtypes,
)
def test_linspace(dt):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dt, q)
    X = dpt.linspace(0, 1, num=2, dtype=dt, sycl_queue=q)
    assert np.allclose(dpt.asnumpy(X), np.linspace(0, 1, num=2, dtype=dt))


def test_linspace_fp():
    q = get_queue_or_skip()
    n = 16
    X = dpt.linspace(0, n - 1, num=n, sycl_queue=q)
    if q.sycl_device.has_aspect_fp64:
        assert X.dtype == dpt.dtype("float64")
    else:
        assert X.dtype == dpt.dtype("float32")
    assert X.shape == (n,)
    assert X.strides == (1,)


@pytest.mark.parametrize("dtype", ["f2", "f4", "f8"])
def test_linspace_fp_max(dtype):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)
    n = 16
    dt = dpt.dtype(dtype)
    max_ = dpt.finfo(dt).max
    X = dpt.linspace(max_, max_, endpoint=True, num=n, dtype=dt, sycl_queue=q)
    assert X.shape == (n,)
    assert X.strides == (1,)
    assert np.allclose(
        dpt.asnumpy(X), np.linspace(max_, max_, endpoint=True, num=n, dtype=dt)
    )


def test_linspace_int():
    q = get_queue_or_skip()
    X = dpt.linspace(0.1, 9.1, 11, endpoint=True, dtype=int, sycl_queue=q)
    Xnp = np.linspace(0.1, 9.1, 11, endpoint=True, dtype=int)
    assert np.array_equal(dpt.asnumpy(X), Xnp)


@pytest.mark.parametrize(
    "dt",
    _all_dtypes,
)
@pytest.mark.parametrize(
    "usm_kind",
    [
        "shared",
        "device",
        "host",
    ],
)
def test_empty_like(dt, usm_kind):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dt, q)

    X = dpt.empty((4, 5), dtype=dt, usm_type=usm_kind, sycl_queue=q)
    Y = dpt.empty_like(X)
    assert X.shape == Y.shape
    assert X.dtype == Y.dtype
    assert X.usm_type == Y.usm_type
    assert X.sycl_queue == Y.sycl_queue

    X = dpt.empty(tuple(), dtype=dt, usm_type=usm_kind, sycl_queue=q)
    Y = dpt.empty_like(X)
    assert X.shape == Y.shape
    assert X.dtype == Y.dtype
    assert X.usm_type == Y.usm_type
    assert X.sycl_queue == Y.sycl_queue


def test_empty_unexpected_data_type():
    with pytest.raises(TypeError):
        dpt.empty(1, dtype=np.object_)


@pytest.mark.parametrize(
    "dt",
    _all_dtypes,
)
@pytest.mark.parametrize(
    "usm_kind",
    [
        "shared",
        "device",
        "host",
    ],
)
def test_zeros_like(dt, usm_kind):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dt, q)

    X = dpt.empty((4, 5), dtype=dt, usm_type=usm_kind, sycl_queue=q)
    Y = dpt.zeros_like(X)
    assert X.shape == Y.shape
    assert X.dtype == Y.dtype
    assert X.usm_type == Y.usm_type
    assert X.sycl_queue == Y.sycl_queue
    assert np.allclose(dpt.asnumpy(Y), np.zeros(X.shape, dtype=X.dtype))

    X = dpt.empty(tuple(), dtype=dt, usm_type=usm_kind, sycl_queue=q)
    Y = dpt.zeros_like(X)
    assert X.shape == Y.shape
    assert X.dtype == Y.dtype
    assert X.usm_type == Y.usm_type
    assert X.sycl_queue == Y.sycl_queue
    assert np.array_equal(dpt.asnumpy(Y), np.zeros(X.shape, dtype=X.dtype))


@pytest.mark.parametrize(
    "dt",
    _all_dtypes,
)
@pytest.mark.parametrize(
    "usm_kind",
    [
        "shared",
        "device",
        "host",
    ],
)
def test_ones_like(dt, usm_kind):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dt, q)

    X = dpt.empty((4, 5), dtype=dt, usm_type=usm_kind, sycl_queue=q)
    Y = dpt.ones_like(X)
    assert X.shape == Y.shape
    assert X.dtype == Y.dtype
    assert X.usm_type == Y.usm_type
    assert X.sycl_queue == Y.sycl_queue
    assert np.allclose(dpt.asnumpy(Y), np.ones(X.shape, dtype=X.dtype))

    X = dpt.empty(tuple(), dtype=dt, usm_type=usm_kind, sycl_queue=q)
    Y = dpt.ones_like(X)
    assert X.shape == Y.shape
    assert X.dtype == Y.dtype
    assert X.usm_type == Y.usm_type
    assert X.sycl_queue == Y.sycl_queue
    assert np.array_equal(dpt.asnumpy(Y), np.ones(X.shape, dtype=X.dtype))


@pytest.mark.parametrize(
    "dt",
    _all_dtypes,
)
@pytest.mark.parametrize(
    "usm_kind",
    [
        "shared",
        "device",
        "host",
    ],
)
def test_full_like(dt, usm_kind):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dt, q)

    fill_v = dpt.dtype(dt).type(1)
    X = dpt.empty((4, 5), dtype=dt, usm_type=usm_kind, sycl_queue=q)
    Y = dpt.full_like(X, fill_v)
    assert X.shape == Y.shape
    assert X.dtype == Y.dtype
    assert X.usm_type == Y.usm_type
    assert X.sycl_queue == Y.sycl_queue
    assert np.allclose(dpt.asnumpy(Y), np.ones(X.shape, dtype=X.dtype))

    X = dpt.empty(tuple(), dtype=dt, usm_type=usm_kind, sycl_queue=q)
    Y = dpt.full_like(X, fill_v)
    assert X.shape == Y.shape
    assert X.dtype == Y.dtype
    assert X.usm_type == Y.usm_type
    assert X.sycl_queue == Y.sycl_queue
    assert np.array_equal(dpt.asnumpy(Y), np.ones(X.shape, dtype=X.dtype))


@pytest.mark.parametrize("dtype", _all_dtypes)
@pytest.mark.parametrize("usm_kind", ["shared", "device", "host"])
def test_eye(dtype, usm_kind):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)

    X = dpt.eye(4, 5, k=1, dtype=dtype, usm_type=usm_kind, sycl_queue=q)
    Xnp = np.eye(4, 5, k=1, dtype=dtype)
    assert X.dtype == Xnp.dtype
    assert np.array_equal(Xnp, dpt.asnumpy(X))


@pytest.mark.parametrize("dtype", _all_dtypes[1:])
def test_tril(dtype):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)

    shape = (2, 3, 4, 5, 5)
    X = dpt.reshape(
        dpt.arange(np.prod(shape), dtype=dtype, sycl_queue=q), shape
    )
    Y = dpt.tril(X)
    Xnp = np.arange(np.prod(shape), dtype=dtype).reshape(shape)
    Ynp = np.tril(Xnp)
    assert Y.dtype == Ynp.dtype
    assert np.array_equal(Ynp, dpt.asnumpy(Y))


@pytest.mark.parametrize("dtype", _all_dtypes[1:])
def test_triu(dtype):
    q = get_queue_or_skip()
    skip_if_dtype_not_supported(dtype, q)

    shape = (4, 5)
    X = dpt.reshape(
        dpt.arange(np.prod(shape), dtype=dtype, sycl_queue=q), shape
    )
    Y = dpt.triu(X, 1)
    Xnp = np.arange(np.prod(shape), dtype=dtype).reshape(shape)
    Ynp = np.triu(Xnp, 1)
    assert Y.dtype == Ynp.dtype
    assert np.array_equal(Ynp, dpt.asnumpy(Y))


@pytest.mark.parametrize("tri_fn", [dpt.tril, dpt.triu])
@pytest.mark.parametrize("usm_type", ["device", "shared", "host"])
def test_tri_usm_type(tri_fn, usm_type):
    q = get_queue_or_skip()
    dtype = dpt.uint16

    shape = (2, 3, 4, 5, 5)
    size = np.prod(shape)
    X = dpt.reshape(
        dpt.arange(size, dtype=dtype, usm_type=usm_type, sycl_queue=q), shape
    )
    Y = tri_fn(X)  # main execution branch
    assert Y.usm_type == X.usm_type
    assert Y.sycl_queue == q
    Y = tri_fn(X, k=-6)  # special case of Y == X
    assert Y.usm_type == X.usm_type
    assert Y.sycl_queue == q
    Y = tri_fn(X, k=6)  # special case of Y == 0
    assert Y.usm_type == X.usm_type
    assert Y.sycl_queue == q


def test_tril_slice():
    q = get_queue_or_skip()

    shape = (6, 10)
    X = dpt.reshape(
        dpt.arange(np.prod(shape), dtype="int", sycl_queue=q), shape
    )[1:, ::-2]
    Y = dpt.tril(X)
    Xnp = np.arange(np.prod(shape), dtype="int").reshape(shape)[1:, ::-2]
    Ynp = np.tril(Xnp)
    assert Y.dtype == Ynp.dtype
    assert np.array_equal(Ynp, dpt.asnumpy(Y))


def test_triu_permute_dims():
    q = get_queue_or_skip()

    shape = (2, 3, 4, 5)
    X = dpt.permute_dims(
        dpt.reshape(
            dpt.arange(np.prod(shape), dtype="int", sycl_queue=q), shape
        ),
        (3, 2, 1, 0),
    )
    Y = dpt.triu(X)
    Xnp = np.transpose(
        np.arange(np.prod(shape), dtype="int").reshape(shape), (3, 2, 1, 0)
    )
    Ynp = np.triu(Xnp)
    assert Y.dtype == Ynp.dtype
    assert np.array_equal(Ynp, dpt.asnumpy(Y))


def test_tril_broadcast_to():
    q = get_queue_or_skip()

    shape = (5, 5)
    X = dpt.broadcast_to(dpt.ones((1), dtype="int", sycl_queue=q), shape)
    Y = dpt.tril(X)
    Xnp = np.broadcast_to(np.ones((1), dtype="int"), shape)
    Ynp = np.tril(Xnp)
    assert Y.dtype == Ynp.dtype
    assert np.array_equal(Ynp, dpt.asnumpy(Y))


def test_triu_bool():
    q = get_queue_or_skip()

    shape = (4, 5)
    X = dpt.ones((shape), dtype="bool", sycl_queue=q)
    Y = dpt.triu(X)
    Xnp = np.ones((shape), dtype="bool")
    Ynp = np.triu(Xnp)
    assert Y.dtype == Ynp.dtype
    assert np.array_equal(Ynp, dpt.asnumpy(Y))


@pytest.mark.parametrize("order", ["F", "C"])
@pytest.mark.parametrize("k", [-10, -2, -1, 3, 4, 10])
def test_triu_order_k(order, k):
    q = get_queue_or_skip()

    shape = (3, 3)
    X = dpt.reshape(
        dpt.arange(np.prod(shape), dtype="int", sycl_queue=q),
        shape,
        order=order,
    )
    Y = dpt.triu(X, k)
    Xnp = np.arange(np.prod(shape), dtype="int").reshape(shape, order=order)
    Ynp = np.triu(Xnp, k)
    assert Y.dtype == Ynp.dtype
    assert X.flags == Y.flags
    assert np.array_equal(Ynp, dpt.asnumpy(Y))


@pytest.mark.parametrize("order", ["F", "C"])
@pytest.mark.parametrize("k", [-10, -4, -3, 1, 2, 10])
def test_tril_order_k(order, k):
    try:
        q = dpctl.SyclQueue()
    except dpctl.SyclQueueCreationError:
        pytest.skip("Queue could not be created")
    shape = (3, 3)
    X = dpt.reshape(
        dpt.arange(np.prod(shape), dtype="int", sycl_queue=q),
        shape,
        order=order,
    )
    Y = dpt.tril(X, k)
    Xnp = np.arange(np.prod(shape), dtype="int").reshape(shape, order=order)
    Ynp = np.tril(Xnp, k)
    assert Y.dtype == Ynp.dtype
    assert X.flags == Y.flags
    assert np.array_equal(Ynp, dpt.asnumpy(Y))


def test_meshgrid():
    q = get_queue_or_skip()

    X = dpt.arange(5, sycl_queue=q)
    Y = dpt.arange(3, sycl_queue=q)
    Z = dpt.meshgrid(X, Y)
    Znp = np.meshgrid(dpt.asnumpy(X), dpt.asnumpy(Y))
    n = len(Z)
    assert n == len(Znp)
    for i in range(n):
        assert np.array_equal(dpt.asnumpy(Z[i]), Znp[i])
    assert dpt.meshgrid() == []
    # dimension > 1 must raise ValueError
    with pytest.raises(ValueError):
        dpt.meshgrid(dpt.usm_ndarray((4, 4)))
    # unknown indexing kwarg must raise ValueError
    with pytest.raises(ValueError):
        dpt.meshgrid(X, indexing="ji")
    # input arrays with different data types must raise ValueError
    with pytest.raises(ValueError):
        dpt.meshgrid(X, dpt.asarray(Y, dtype="b1"))


def test_meshgrid2():
    q1 = get_queue_or_skip()
    q2 = get_queue_or_skip()
    q3 = get_queue_or_skip()

    x1 = dpt.arange(0, 2, dtype="int16", sycl_queue=q1)
    x2 = dpt.arange(3, 6, dtype="int16", sycl_queue=q2)
    x3 = dpt.arange(6, 10, dtype="int16", sycl_queue=q3)
    y1, y2, y3 = dpt.meshgrid(x1, x2, x3, indexing="xy")
    z1, z2, z3 = dpt.meshgrid(x1, x2, x3, indexing="ij")
    assert all(
        x.sycl_queue == y.sycl_queue for x, y in zip((x1, x2, x3), (y1, y2, y3))
    )
    assert all(
        x.sycl_queue == z.sycl_queue for x, z in zip((x1, x2, x3), (z1, z2, z3))
    )
    assert y1.shape == y2.shape and y2.shape == y3.shape
    assert z1.shape == z2.shape and z2.shape == z3.shape
    assert y1.shape == (len(x2), len(x1), len(x3))
    assert z1.shape == (len(x1), len(x2), len(x3))
    # FIXME: uncomment out once gh-921 is merged
    # assert all(z.flags["C"]  for z in (z1, z2, z3))
    # assert all(y.flags["C"]  for y in (y1, y2, y3))


def test_common_arg_validation():
    order = "I"
    # invalid order must raise ValueError
    with pytest.raises(ValueError):
        dpt.empty(10, order=order)
    with pytest.raises(ValueError):
        dpt.zeros(10, order=order)
    with pytest.raises(ValueError):
        dpt.ones(10, order=order)
    with pytest.raises(ValueError):
        dpt.full(10, 1, order=order)
    with pytest.raises(ValueError):
        dpt.eye(10, order=order)
    X = dpt.empty(10)
    with pytest.raises(ValueError):
        dpt.empty_like(X, order=order)
    with pytest.raises(ValueError):
        dpt.zeros_like(X, order=order)
    with pytest.raises(ValueError):
        dpt.ones_like(X, order=order)
    with pytest.raises(ValueError):
        dpt.full_like(X, 1, order=order)
    X = dict()
    # test for type validation
    with pytest.raises(TypeError):
        dpt.empty_like(X)
    with pytest.raises(TypeError):
        dpt.zeros_like(X)
    with pytest.raises(TypeError):
        dpt.ones_like(X)
    with pytest.raises(TypeError):
        dpt.full_like(X, 1)
    with pytest.raises(TypeError):
        dpt.tril(X)
    with pytest.raises(TypeError):
        dpt.triu(X)
    with pytest.raises(TypeError):
        dpt.meshgrid(X)


def test_flags():
    x = dpt.empty(tuple(), "i4")
    f = x.flags
    f.__repr__()
    f.c_contiguous
    f.f_contiguous
    f.contiguous
    f.fc
    f.fnc
    f.forc
    f.writable
    # check comparison with generic types
    f == Ellipsis


def test_asarray_uint64():
    Xnp = np.ndarray(1, dtype=np.uint64)
    X = dpt.asarray(Xnp)
    assert X.dtype == Xnp.dtype


def test_Device():
    try:
        dev = dpctl.select_default_device()
        d1 = dpt.Device.create_device(dev)
        d2 = dpt.Device.create_device(dev)
    except (dpctl.SyclQueueCreationError, dpctl.SyclDeviceCreationError):
        pytest.skip(
            "Could not create default device, or a queue that targets it"
        )
    assert d1 == d2
    dict = {d1: 1}
    assert dict[d2] == 1
    assert d1 == d2.sycl_queue
    assert not d1 == Ellipsis



#                       Data Parallel Control (dpctl)
#
#  Copyright 2020-2022 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import operator

import numpy as np
from numpy.core.numeric import normalize_axis_index

import dpctl
import dpctl.tensor as dpt
import dpctl.tensor._tensor_impl as ti

from ._copy_utils import _extract_impl, _nonzero_impl


def take(x, indices, /, *, axis=None, mode="clip"):
    """take(x, indices, axis=None, mode="clip")

    Takes elements from array along a given axis.

    Args:
       x: usm_ndarray
          The array that elements will be taken from.
       indices: usm_ndarray
          One-dimensional array of indices.
       axis:
          The axis over which the values will be selected.
          If x is one-dimensional, this argument is optional.
       mode:
          How out-of-bounds indices will be handled.
          "Clip" - clamps indices to (-n <= i < n), then wraps
          negative indices.
          "Wrap" - wraps both negative and positive indices.

    Returns:
       out: usm_ndarray
          Array with shape x.shape[:axis] + indices.shape + x.shape[axis + 1:]
          filled with elements .
    """
    if not isinstance(x, dpt.usm_ndarray):
        raise TypeError(
            "Expected instance of `dpt.usm_ndarray`, got `{}`.".format(type(x))
        )

    if not isinstance(indices, dpt.usm_ndarray):
        raise TypeError(
            "`indices` expected `dpt.usm_ndarray`, got `{}`.".format(
                type(indices)
            )
        )
    if not np.issubdtype(indices.dtype, np.integer):
        raise IndexError(
            "`indices` expected integer data type, got `{}`".format(
                indices.dtype
            )
        )
    if indices.ndim != 1:
        raise ValueError(
            "`indices` expected a 1D array, got `{}`".format(indices.ndim)
        )
    exec_q = dpctl.utils.get_execution_queue([x.sycl_queue, indices.sycl_queue])
    if exec_q is None:
        raise dpctl.utils.ExecutionPlacementError
    res_usm_type = dpctl.utils.get_coerced_usm_type(
        [x.usm_type, indices.usm_type]
    )

    modes = {"clip": 0, "wrap": 1}
    try:
        mode = modes[mode]
    except KeyError:
        raise ValueError("`mode` must be `clip` or `wrap`.")

    x_ndim = x.ndim
    if axis is None:
        if x_ndim > 1:
            raise ValueError(
                "`axis` cannot be `None` for array of dimension `{}`".format(
                    x_ndim
                )
            )
        axis = 0

    if x_ndim > 0:
        axis = normalize_axis_index(operator.index(axis), x_ndim)
        res_shape = x.shape[:axis] + indices.shape + x.shape[axis + 1 :]
    else:
        if axis != 0:
            raise ValueError("`axis` must be 0 for an array of dimension 0.")
        res_shape = indices.shape

    res = dpt.empty(
        res_shape, dtype=x.dtype, usm_type=res_usm_type, sycl_queue=exec_q
    )

    hev, _ = ti._take(x, (indices,), res, axis, mode, sycl_queue=exec_q)
    hev.wait()

    return res


def put(x, indices, vals, /, *, axis=None, mode="clip"):
    """put(x, indices, vals, axis=None, mode="clip")

    Puts values of an array into another array
    along a given axis.

    Args:
       x: usm_ndarray
          The array the values will be put into.
       indices: usm_ndarray
          One-dimensional array of indices.
       vals:
          Array of values to be put into `x`.
          Must be broadcastable to the shape of `indices`.
       axis:
          The axis over which the values will be placed.
          If x is one-dimensional, this argument is optional.
       mode:
          How out-of-bounds indices will be handled.
          "Clip" - clamps indices to (-axis_size <= i < axis_size),
          then wraps negative indices.
          "Wrap" - wraps both negative and positive indices.
    """
    if not isinstance(x, dpt.usm_ndarray):
        raise TypeError(
            "Expected instance of `dpt.usm_ndarray`, got `{}`.".format(type(x))
        )
    if isinstance(vals, dpt.usm_ndarray):
        queues_ = [x.sycl_queue, vals.sycl_queue]
        usm_types_ = [x.usm_type, vals.usm_type]
    else:
        queues_ = [
            x.sycl_queue,
        ]
        usm_types_ = [
            x.usm_type,
        ]
    if not isinstance(indices, dpt.usm_ndarray):
        raise TypeError(
            "`indices` expected `dpt.usm_ndarray`, got `{}`.".format(
                type(indices)
            )
        )
    if indices.ndim != 1:
        raise ValueError(
            "`indices` expected a 1D array, got `{}`".format(indices.ndim)
        )
    if not np.issubdtype(indices.dtype, np.integer):
        raise IndexError(
            "`indices` expected integer data type, got `{}`".format(
                indices.dtype
            )
        )
    queues_.append(indices.sycl_queue)
    usm_types_.append(indices.usm_type)
    exec_q = dpctl.utils.get_execution_queue(queues_)
    if exec_q is None:
        raise dpctl.utils.ExecutionPlacementError
    vals_usm_type = dpctl.utils.get_coerced_usm_type(usm_types_)
    modes = {"clip": 0, "wrap": 1}
    try:
        mode = modes[mode]
    except KeyError:
        raise ValueError("`mode` must be `clip` or `wrap`.")

    x_ndim = x.ndim
    if axis is None:
        if x_ndim > 1:
            raise ValueError(
                "`axis` cannot be `None` for array of dimension `{}`".format(
                    x_ndim
                )
            )
        axis = 0

    if x_ndim > 0:
        axis = normalize_axis_index(operator.index(axis), x_ndim)

        val_shape = x.shape[:axis] + indices.shape + x.shape[axis + 1 :]
    else:
        if axis != 0:
            raise ValueError("`axis` must be 0 for an array of dimension 0.")
        val_shape = indices.shape

    if not isinstance(vals, dpt.usm_ndarray):
        vals = dpt.asarray(
            vals, dtype=x.dtype, usm_type=vals_usm_type, sycl_queue=exec_q
        )

    vals = dpt.broadcast_to(vals, val_shape)

    hev, _ = ti._put(x, (indices,), vals, axis, mode, sycl_queue=exec_q)
    hev.wait()


def extract(condition, arr):
    """extract(condition, arr)

    Returns the elements of an array that satisfies the condition.

    If `condition` is boolean ``dpctl.tensor.extract`` is
    equivalent to ``arr[condition]``.

    Note that ``dpctl.tensor.place`` does the opposite of
    ``dpctl.tensor.extract``.

    Args:
       conditions: usm_ndarray
            An array whose non-zero or True entries indicate the element
            of `arr` to extract.
       arr: usm_ndarray
            Input array of the same size as `condition`.

    Returns:
        usm_ndarray
            Rank 1 array of values from `arr` where `condition` is True.
    """
    if not isinstance(condition, dpt.usm_ndarray):
        raise TypeError(
            "Expecting dpctl.tensor.usm_ndarray type, " f"got {type(condition)}"
        )
    if not isinstance(arr, dpt.usm_ndarray):
        raise TypeError(
            "Expecting dpctl.tensor.usm_ndarray type, " f"got {type(arr)}"
        )
    exec_q = dpctl.utils.get_execution_queue(
        (
            condition.sycl_queue,
            arr.sycl_queue,
        )
    )
    if exec_q is None:
        raise dpctl.utils.ExecutionPlacementError
    if condition.shape != arr.shape:
        raise ValueError("Arrays are not of the same size")
    return _extract_impl(arr, condition)


def place(arr, mask, vals):
    """place(arr, mask, vals)

    Change elements of an array based on conditional and input values.

    If `mask` is boolean ``dpctl.tensor.place`` is
    equivalent to ``arr[condition] = vals``.

    Args:
        arr: usm_ndarray
            Array to put data into.
       mask: usm_ndarray
            Boolean mask array. Must have the same size as `arr`.
       vals: usm_ndarray
            Values to put into `arr`. Only the first N elements are
            used, where N is the number of True values in `mask`. If
            `vals` is smaller than N, it will be repeated, and if
            elements of `arr` are to be masked, this sequence must be
            non-empty. Array `vals` must be one dimensional.
    """
    if not isinstance(arr, dpt.usm_ndarray):
        raise TypeError(
            "Expecting dpctl.tensor.usm_ndarray type, " f"got {type(arr)}"
        )
    if not isinstance(mask, dpt.usm_ndarray):
        raise TypeError(
            "Expecting dpctl.tensor.usm_ndarray type, " f"got {type(mask)}"
        )
    if not isinstance(vals, dpt.usm_ndarray):
        raise TypeError(
            "Expecting dpctl.tensor.usm_ndarray type, " f"got {type(vals)}"
        )
    exec_q = dpctl.utils.get_execution_queue(
        (
            arr.sycl_queue,
            mask.sycl_queue,
            vals.sycl_queue,
        )
    )
    if exec_q is None:
        raise dpctl.utils.ExecutionPlacementError
    if arr.shape != mask.shape or vals.ndim != 1:
        raise ValueError("Array sizes are not as required")
    cumsum = dpt.empty(mask.size, dtype="i8", sycl_queue=exec_q)
    nz_count = ti.mask_positions(mask, cumsum, sycl_queue=exec_q)
    if nz_count == 0:
        return
    if vals.dtype == arr.dtype:
        rhs = vals
    else:
        rhs = dpt.astype(vals, arr.dtype)
    hev, _ = ti._place(
        dst=arr,
        cumsum=cumsum,
        axis_start=0,
        axis_end=mask.ndim,
        rhs=rhs,
        sycl_queue=exec_q,
    )
    hev.wait()


def nonzero(arr):
    """nonzero(arr)

    Return the indices of non-zero elements.

    Returns the tuple of usm_narrays, one for each dimension
    of `arr`, containing the indices of the non-zero elements
    in that dimension. The values of `arr` are always tested in
    row-major, C-style order.

    Args:
        arr: usm_ndarray
            Input array, which has non-zero array rank.
    Returns:
        Tuple[usm_ndarray]
            Indices of non-zero array elements.
    """
    if not isinstance(arr, dpt.usm_ndarray):
        raise TypeError(
            "Expecting dpctl.tensor.usm_ndarray type, " f"got {type(arr)}"
        )
    if arr.ndim == 0:
        raise ValueError("Array of positive rank is exepcted")
    return _nonzero_impl(arr)



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# cython: language=c++
# cython: language_level=3

cimport dpctl as c_dpctl

import dpctl


def call_create_from_context_and_devices():
    cdef c_dpctl.SyclQueue q
    d = dpctl.SyclDevice()
    ctx = dpctl.SyclContext(d)
    # calling static method
    q = c_dpctl.SyclQueue._create_from_context_and_device(
        <c_dpctl.SyclContext> ctx,
        <c_dpctl.SyclDevice> d
    )
    return q




# This file was generated by 'versioneer.py' (0.28) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-07T14:13:01-0600",
 "dirty": false,
 "error": null,
 "full-revisionid": "6dc5479eca87556eb9eb1100bc0cd0f6bc9d3a13",
 "version": "0.14.2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



#                       Data Parallel Control (dpctl)
#
#  Copyright 2020-2022 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import operator

import numpy as np
from numpy.core.numeric import normalize_axis_index

import dpctl
import dpctl.memory as dpm
import dpctl.tensor as dpt
import dpctl.tensor._tensor_impl as ti
import dpctl.utils
from dpctl.tensor._device import normalize_queue_device

__doc__ = (
    "Implementation module for copy- and cast- operations on "
    ":class:`dpctl.tensor.usm_ndarray`."
)


def _has_memory_overlap(x1, x2):
    if x1.size and x2.size:
        m1 = dpm.as_usm_memory(x1)
        m2 = dpm.as_usm_memory(x2)
        # can only overlap if bound to the same context
        if m1.sycl_context == m2.sycl_context:
            p1_beg = m1._pointer
            p1_end = p1_beg + m1.nbytes
            p2_beg = m2._pointer
            p2_end = p2_beg + m2.nbytes
            # may intersect if not ((p1_beg >= p2_end) or (p2_beg >= p2_end))
            return (p1_beg < p2_end) and (p2_beg < p1_end)
        return False
    # zero element array do not overlap anything
    return False


def _copy_to_numpy(ary):
    if not isinstance(ary, dpt.usm_ndarray):
        raise TypeError
    nb = ary.usm_data.nbytes
    hh = dpm.MemoryUSMHost(nb, queue=ary.sycl_queue)
    hh.copy_from_device(ary.usm_data)
    h = np.ndarray(nb, dtype="u1", buffer=hh).view(ary.dtype)
    itsz = ary.itemsize
    strides_bytes = tuple(si * itsz for si in ary.strides)
    offset = ary.__sycl_usm_array_interface__.get("offset", 0) * itsz
    return np.ndarray(
        ary.shape,
        dtype=ary.dtype,
        buffer=h,
        strides=strides_bytes,
        offset=offset,
    )


def _copy_from_numpy(np_ary, usm_type="device", sycl_queue=None):
    "Copies numpy array `np_ary` into a new usm_ndarray"
    # This may peform a copy to meet stated requirements
    Xnp = np.require(np_ary, requirements=["A", "E"])
    alloc_q = normalize_queue_device(sycl_queue=sycl_queue, device=None)
    dt = Xnp.dtype
    if dt.char in "dD" and alloc_q.sycl_device.has_aspect_fp64 is False:
        Xusm_dtype = (
            dpt.dtype("float32") if dt.char == "d" else dpt.dtype("complex64")
        )
    else:
        Xusm_dtype = dt
    Xusm = dpt.empty(
        Xnp.shape, dtype=Xusm_dtype, usm_type=usm_type, sycl_queue=sycl_queue
    )
    _copy_from_numpy_into(Xusm, Xnp)
    return Xusm


def _copy_from_numpy_into(dst, np_ary):
    "Copies `np_ary` into `dst` of type :class:`dpctl.tensor.usm_ndarray"
    if not isinstance(np_ary, np.ndarray):
        raise TypeError(f"Expected numpy.ndarray, got {type(np_ary)}")
    if not isinstance(dst, dpt.usm_ndarray):
        raise TypeError(f"Expected usm_ndarray, got {type(dst)}")
    src_ary = np.broadcast_to(np_ary, dst.shape)
    copy_q = dst.sycl_queue
    if copy_q.sycl_device.has_aspect_fp64 is False:
        src_ary_dt_c = src_ary.dtype.char
        if src_ary_dt_c == "d":
            src_ary = src_ary.astype(np.float32)
        elif src_ary_dt_c == "D":
            src_ary = src_ary.astype(np.complex64)
    ti._copy_numpy_ndarray_into_usm_ndarray(
        src=src_ary, dst=dst, sycl_queue=copy_q
    )


def from_numpy(np_ary, device=None, usm_type="device", sycl_queue=None):
    """
    from_numpy(arg, device=None, usm_type="device", sycl_queue=None)

    Creates :class:`dpctl.tensor.usm_ndarray` from instance of
    `numpy.ndarray`.

    Args:
        arg: An instance of `numpy.ndarray`
        device: array API specification of device where the output array
            is created.
        sycl_queue: a :class:`dpctl.SyclQueue` used to create the output
            array is created
    """
    q = normalize_queue_device(sycl_queue=sycl_queue, device=device)
    return _copy_from_numpy(np_ary, usm_type=usm_type, sycl_queue=q)


def to_numpy(usm_ary):
    """
    to_numpy(usm_ary)

    Copies content of :class:`dpctl.tensor.usm_ndarray` instance `usm_ary`
    into `numpy.ndarray` instance of the same shape and same data type.

    Args:
        usm_ary: An instance of :class:`dpctl.tensor.usm_ndarray`
    Returns:
        An instance of `numpy.ndarray` populated with content of `usm_ary`.
    """
    return _copy_to_numpy(usm_ary)


def asnumpy(usm_ary):
    """
    asnumpy(usm_ary)

    Copies content of :class:`dpctl.tensor.usm_ndarray` instance `usm_ary`
    into `numpy.ndarray` instance of the same shape and same data type.

    Args:
        usm_ary: An instance of :class:`dpctl.tensor.usm_ndarray`
    Returns:
        An instance of `numpy.ndarray` populated with content of `usm_ary`.
    """
    return _copy_to_numpy(usm_ary)


class Dummy:
    "Helper class with specified __sycl_usm_array_interface__ attribute"

    def __init__(self, iface):
        self.__sycl_usm_array_interface__ = iface


def _copy_overlapping(dst, src):
    """Assumes src and dst have the same shape."""
    q = normalize_queue_device(sycl_queue=dst.sycl_queue)
    tmp = dpt.usm_ndarray(
        src.shape,
        dtype=src.dtype,
        buffer="device",
        order="C",
        buffer_ctor_kwargs={"queue": q},
    )
    hcp1, cp1 = ti._copy_usm_ndarray_into_usm_ndarray(
        src=src, dst=tmp, sycl_queue=q
    )
    hcp2, _ = ti._copy_usm_ndarray_into_usm_ndarray(
        src=tmp, dst=dst, sycl_queue=q, depends=[cp1]
    )
    hcp2.wait()
    hcp1.wait()


def _copy_same_shape(dst, src):
    """Assumes src and dst have the same shape."""
    # check that memory regions do not overlap
    if _has_memory_overlap(dst, src):
        _copy_overlapping(src=src, dst=dst)
        return

    hev, _ = ti._copy_usm_ndarray_into_usm_ndarray(
        src=src, dst=dst, sycl_queue=dst.sycl_queue
    )
    hev.wait()


if hasattr(np, "broadcast_shapes"):

    def _broadcast_shapes(sh1, sh2):
        return np.broadcast_shapes(sh1, sh2)

else:

    def _broadcast_shapes(sh1, sh2):
        # use arrays with zero strides, whose memory footprint
        # is independent of the number of array elements
        return np.broadcast(
            np.empty(sh1, dtype=[]),
            np.empty(sh2, dtype=[]),
        ).shape


def _copy_from_usm_ndarray_to_usm_ndarray(dst, src):
    if any(
        not isinstance(arg, dpt.usm_ndarray)
        for arg in (
            dst,
            src,
        )
    ):
        raise TypeError(
            "Both types are expected to be dpctl.tensor.usm_ndarray, "
            f"got {type(dst)} and {type(src)}."
        )

    if dst.ndim == src.ndim and dst.shape == src.shape:
        _copy_same_shape(dst, src)
        return

    try:
        common_shape = _broadcast_shapes(dst.shape, src.shape)
    except ValueError as exc:
        raise ValueError("Shapes of two arrays are not compatible") from exc

    if dst.size < src.size:
        raise ValueError("Destination is smaller ")

    if len(common_shape) > dst.ndim:
        ones_count = len(common_shape) - dst.ndim
        for k in range(ones_count):
            if common_shape[k] != 1:
                raise ValueError
        common_shape = common_shape[ones_count:]

    if src.ndim < len(common_shape):
        new_src_strides = (0,) * (len(common_shape) - src.ndim) + src.strides
        src_same_shape = dpt.usm_ndarray(
            common_shape, dtype=src.dtype, buffer=src, strides=new_src_strides
        )
    else:
        src_same_shape = src
        src_same_shape.shape = common_shape

    _copy_same_shape(dst, src_same_shape)


def copy(usm_ary, order="K"):
    """
    Creates a copy of given instance of `usm_ndarray`.

    Memory layour of the copy is controlled by `order` keyword,
    following NumPy's conventions. The `order` keywords can be
    one of the following:

       - "C": C-contiguous memory layout
       - "F": Fortran-contiguous memory layout
       - "A": Fortran-contiguous if the input array is also Fortran-contiguous,
         otherwise C-contiguous
       - "K": match the layout of `usm_ary` as closely as possible.

    """
    if not isinstance(usm_ary, dpt.usm_ndarray):
        return TypeError(
            f"Expected object of type dpt.usm_ndarray, got {type(usm_ary)}"
        )
    copy_order = "C"
    if order == "C":
        pass
    elif order == "F":
        copy_order = order
    elif order == "A":
        if usm_ary.flags.f_contiguous:
            copy_order = "F"
    elif order == "K":
        if usm_ary.flags.f_contiguous:
            copy_order = "F"
    else:
        raise ValueError(
            "Unrecognized value of the order keyword. "
            "Recognized values are 'A', 'C', 'F', or 'K'"
        )
    c_contig = usm_ary.flags.c_contiguous
    f_contig = usm_ary.flags.f_contiguous
    R = dpt.usm_ndarray(
        usm_ary.shape,
        dtype=usm_ary.dtype,
        buffer=usm_ary.usm_type,
        order=copy_order,
        buffer_ctor_kwargs={"queue": usm_ary.sycl_queue},
    )
    if order == "K" and (not c_contig and not f_contig):
        original_strides = usm_ary.strides
        ind = sorted(
            range(usm_ary.ndim),
            key=lambda i: abs(original_strides[i]),
            reverse=True,
        )
        new_strides = tuple(R.strides[ind[i]] for i in ind)
        R = dpt.usm_ndarray(
            usm_ary.shape,
            dtype=usm_ary.dtype,
            buffer=R.usm_data,
            strides=new_strides,
        )
    _copy_same_shape(R, usm_ary)
    return R


def astype(usm_ary, newdtype, order="K", casting="unsafe", copy=True):
    """
    astype(usm_array, new_dtype, order="K", casting="unsafe", copy=True)

    Returns a copy of the array, cast to a specified type.

    A view can be returned, if possible, when `copy=False` is used.
    """
    if not isinstance(usm_ary, dpt.usm_ndarray):
        return TypeError(
            f"Expected object of type dpt.usm_ndarray, got {type(usm_ary)}"
        )
    if not isinstance(order, str) or order not in ["A", "C", "F", "K"]:
        raise ValueError(
            "Unrecognized value of the order keyword. "
            "Recognized values are 'A', 'C', 'F', or 'K'"
        )
    ary_dtype = usm_ary.dtype
    target_dtype = dpt.dtype(newdtype)
    if not dpt.can_cast(ary_dtype, target_dtype, casting=casting):
        raise TypeError(
            f"Can not cast from {ary_dtype} to {newdtype} "
            f"according to rule {casting}."
        )
    c_contig = usm_ary.flags.c_contiguous
    f_contig = usm_ary.flags.f_contiguous
    needs_copy = copy or not ary_dtype == target_dtype
    if not needs_copy and (order != "K"):
        needs_copy = (c_contig and order not in ["A", "C"]) or (
            f_contig and order not in ["A", "F"]
        )
    if not needs_copy:
        return usm_ary
    copy_order = "C"
    if order == "C":
        pass
    elif order == "F":
        copy_order = order
    elif order == "A":
        if usm_ary.flags.f_contiguous:
            copy_order = "F"
    elif order == "K":
        if usm_ary.flags.f_contiguous:
            copy_order = "F"
    else:
        raise ValueError(
            "Unrecognized value of the order keyword. "
            "Recognized values are 'A', 'C', 'F', or 'K'"
        )
    R = dpt.usm_ndarray(
        usm_ary.shape,
        dtype=target_dtype,
        buffer=usm_ary.usm_type,
        order=copy_order,
        buffer_ctor_kwargs={"queue": usm_ary.sycl_queue},
    )
    if order == "K" and (not c_contig and not f_contig):
        original_strides = usm_ary.strides
        ind = sorted(
            range(usm_ary.ndim),
            key=lambda i: abs(original_strides[i]),
            reverse=True,
        )
        new_strides = tuple(R.strides[ind[i]] for i in ind)
        R = dpt.usm_ndarray(
            usm_ary.shape,
            dtype=target_dtype,
            buffer=R.usm_data,
            strides=new_strides,
        )
    _copy_from_usm_ndarray_to_usm_ndarray(R, usm_ary)
    return R


def _extract_impl(ary, ary_mask, axis=0):
    """Extract elements of ary by applying mask starting from slot
    dimension axis"""
    if not isinstance(ary, dpt.usm_ndarray):
        raise TypeError(
            f"Expecting type dpctl.tensor.usm_ndarray, got {type(ary)}"
        )
    if not isinstance(ary_mask, dpt.usm_ndarray):
        raise TypeError(
            f"Expecting type dpctl.tensor.usm_ndarray, got {type(ary_mask)}"
        )
    exec_q = dpctl.utils.get_execution_queue(
        (ary.sycl_queue, ary_mask.sycl_queue)
    )
    if exec_q is None:
        raise dpctl.utils.ExecutionPlacementError(
            "arrays have different associated queues. "
            "Use `Y.to_device(X.device)` to migrate."
        )
    ary_nd = ary.ndim
    pp = normalize_axis_index(operator.index(axis), ary_nd)
    mask_nd = ary_mask.ndim
    if pp < 0 or pp + mask_nd > ary_nd:
        raise ValueError(
            "Parameter p is inconsistent with input array dimensions"
        )
    mask_nelems = ary_mask.size
    cumsum = dpt.empty(mask_nelems, dtype=dpt.int64, device=ary_mask.device)
    exec_q = cumsum.sycl_queue
    mask_count = ti.mask_positions(ary_mask, cumsum, sycl_queue=exec_q)
    dst_shape = ary.shape[:pp] + (mask_count,) + ary.shape[pp + mask_nd :]
    dst = dpt.empty(
        dst_shape, dtype=ary.dtype, usm_type=ary.usm_type, device=ary.device
    )
    hev, _ = ti._extract(
        src=ary,
        cumsum=cumsum,
        axis_start=pp,
        axis_end=pp + mask_nd,
        dst=dst,
        sycl_queue=exec_q,
    )
    hev.wait()
    return dst


def _nonzero_impl(ary):
    if not isinstance(ary, dpt.usm_ndarray):
        raise TypeError(
            f"Expecting type dpctl.tensor.usm_ndarray, got {type(ary)}"
        )
    exec_q = ary.sycl_queue
    usm_type = ary.usm_type
    mask_nelems = ary.size
    cumsum = dpt.empty(
        mask_nelems, dtype=dpt.int64, sycl_queue=exec_q, order="C"
    )
    mask_count = ti.mask_positions(ary, cumsum, sycl_queue=exec_q)
    indexes = dpt.empty(
        (ary.ndim, mask_count),
        dtype=cumsum.dtype,
        usm_type=usm_type,
        sycl_queue=exec_q,
        order="C",
    )
    hev, _ = ti._nonzero(cumsum, indexes, ary.shape, exec_q)
    res = tuple(indexes[i, :] for i in range(ary.ndim))
    hev.wait()
    return res


def _take_multi_index(ary, inds, p):
    if not isinstance(ary, dpt.usm_ndarray):
        raise TypeError
    queues_ = [
        ary.sycl_queue,
    ]
    usm_types_ = [
        ary.usm_type,
    ]
    if not isinstance(inds, list) and not isinstance(inds, tuple):
        inds = (inds,)
    all_integers = True
    for ind in inds:
        queues_.append(ind.sycl_queue)
        usm_types_.append(ind.usm_type)
        if all_integers:
            all_integers = ind.dtype.kind in "ui"
    exec_q = dpctl.utils.get_execution_queue(queues_)
    if exec_q is None:
        raise dpctl.utils.ExecutionPlacementError("")
    if not all_integers:
        raise IndexError(
            "arrays used as indices must be of integer (or boolean) type"
        )
    if len(inds) > 1:
        inds = dpt.broadcast_arrays(*inds)
    ary_ndim = ary.ndim
    p = normalize_axis_index(operator.index(p), ary_ndim)

    res_shape = ary.shape[:p] + inds[0].shape + ary.shape[p + len(inds) :]
    res_usm_type = dpctl.utils.get_coerced_usm_type(usm_types_)
    res = dpt.empty(
        res_shape, dtype=ary.dtype, usm_type=res_usm_type, sycl_queue=exec_q
    )

    hev, _ = ti._take(
        src=ary, ind=inds, dst=res, axis_start=p, mode=0, sycl_queue=exec_q
    )
    hev.wait()

    return res


def _place_impl(ary, ary_mask, vals, axis=0):
    """Extract elements of ary by applying mask starting from slot
    dimension axis"""
    if not isinstance(ary, dpt.usm_ndarray):
        raise TypeError(
            f"Expecting type dpctl.tensor.usm_ndarray, got {type(ary)}"
        )
    if not isinstance(ary_mask, dpt.usm_ndarray):
        raise TypeError(
            f"Expecting type dpctl.tensor.usm_ndarray, got {type(ary_mask)}"
        )
    exec_q = dpctl.utils.get_execution_queue(
        (
            ary.sycl_queue,
            ary_mask.sycl_queue,
        )
    )
    if exec_q is not None:
        if not isinstance(vals, dpt.usm_ndarray):
            vals = dpt.asarray(vals, dtype=ary.dtype, sycl_queue=exec_q)
        else:
            exec_q = dpctl.utils.get_execution_queue((exec_q, vals.sycl_queue))
    if exec_q is None:
        raise dpctl.utils.ExecutionPlacementError(
            "arrays have different associated queues. "
            "Use `Y.to_device(X.device)` to migrate."
        )
    ary_nd = ary.ndim
    pp = normalize_axis_index(operator.index(axis), ary_nd)
    mask_nd = ary_mask.ndim
    if pp < 0 or pp + mask_nd > ary_nd:
        raise ValueError(
            "Parameter p is inconsistent with input array dimensions"
        )
    mask_nelems = ary_mask.size
    cumsum = dpt.empty(mask_nelems, dtype=dpt.int64, device=ary_mask.device)
    exec_q = cumsum.sycl_queue
    mask_count = ti.mask_positions(ary_mask, cumsum, sycl_queue=exec_q)
    expected_vals_shape = (
        ary.shape[:pp] + (mask_count,) + ary.shape[pp + mask_nd :]
    )
    if vals.dtype == ary.dtype:
        rhs = vals
    else:
        rhs = dpt.astype(vals, ary.dtype)
    rhs = dpt.broadcast_to(rhs, expected_vals_shape)
    hev, _ = ti._place(
        dst=ary,
        cumsum=cumsum,
        axis_start=pp,
        axis_end=pp + mask_nd,
        rhs=rhs,
        sycl_queue=exec_q,
    )
    hev.wait()
    return


def _put_multi_index(ary, inds, p, vals):
    if isinstance(vals, dpt.usm_ndarray):
        queues_ = [ary.sycl_queue, vals.sycl_queue]
        usm_types_ = [ary.usm_type, vals.usm_type]
    else:
        queues_ = [
            ary.sycl_queue,
        ]
        usm_types_ = [
            ary.usm_type,
        ]
    if not isinstance(inds, list) and not isinstance(inds, tuple):
        inds = (inds,)
    all_integers = True
    for ind in inds:
        if not isinstance(ind, dpt.usm_ndarray):
            raise TypeError
        queues_.append(ind.sycl_queue)
        usm_types_.append(ind.usm_type)
        if all_integers:
            all_integers = ind.dtype.kind in "ui"
    exec_q = dpctl.utils.get_execution_queue(queues_)
    if exec_q is None:
        raise dpctl.utils.ExecutionPlacementError(
            "Can not automatically determine where to allocate the "
            "result or performance execution. "
            "Use `usm_ndarray.to_device` method to migrate data to "
            "be associated with the same queue."
        )
    if not all_integers:
        raise IndexError(
            "arrays used as indices must be of integer (or boolean) type"
        )
    if len(inds) > 1:
        inds = dpt.broadcast_arrays(*inds)
    ary_ndim = ary.ndim

    p = normalize_axis_index(operator.index(p), ary_ndim)
    vals_shape = ary.shape[:p] + inds[0].shape + ary.shape[p + len(inds) :]

    vals_usm_type = dpctl.utils.get_coerced_usm_type(usm_types_)
    if not isinstance(vals, dpt.usm_ndarray):
        vals = dpt.asarray(
            vals, ary.dtype, usm_type=vals_usm_type, sycl_queue=exec_q
        )

    vals = dpt.broadcast_to(vals, vals_shape)

    hev, _ = ti._put(
        dst=ary, ind=inds, val=vals, axis_start=p, mode=0, sycl_queue=exec_q
    )
    hev.wait()

    return



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
from helper import get_queue_or_skip, skip_if_dtype_not_supported

import dpctl.tensor as dpt


class TestPrint:
    def setup_method(self):
        self._retain_options = dpt.get_print_options()

    def teardown_method(self):
        dpt.set_print_options(**self._retain_options)


class TestArgValidation(TestPrint):
    @pytest.mark.parametrize(
        "arg,err",
        [
            ({"linewidth": "I"}, TypeError),
            ({"edgeitems": "I"}, TypeError),
            ({"threshold": "I"}, TypeError),
            ({"precision": "I"}, TypeError),
            ({"floatmode": "I"}, ValueError),
            ({"edgeitems": "I"}, TypeError),
            ({"sign": "I"}, ValueError),
            ({"nanstr": np.nan}, TypeError),
            ({"infstr": np.nan}, TypeError),
        ],
    )
    def test_print_option_arg_validation(self, arg, err):
        with pytest.raises(err):
            dpt.set_print_options(**arg)

    def test_usm_ndarray_repr_arg_validation(self):
        X = dict()
        with pytest.raises(TypeError):
            dpt.usm_ndarray_repr(X)

        X = dpt.arange(4)
        with pytest.raises(TypeError):
            dpt.usm_ndarray_repr(X, line_width="I")

        with pytest.raises(TypeError):
            dpt.usm_ndarray_repr(X, precision="I")

        with pytest.raises(TypeError):
            dpt.usm_ndarray_repr(X, prefix=4)

    def test_usm_ndarray_str_arg_validation(self):
        X = dict()
        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X)

        X = dpt.arange(4)
        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X, line_width="I")

        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X, edge_items="I")

        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X, threshold="I")

        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X, precision="I")

        with pytest.raises(ValueError):
            dpt.usm_ndarray_str(X, floatmode="I")

        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X, edge_items="I")

        with pytest.raises(ValueError):
            dpt.usm_ndarray_str(X, sign="I")

        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X, prefix=4)

        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X, prefix=4)

        with pytest.raises(TypeError):
            dpt.usm_ndarray_str(X, suffix=4)


class TestSetPrintOptions(TestPrint):
    def test_set_linewidth(self):
        q = get_queue_or_skip()

        dpt.set_print_options(linewidth=1)
        x = dpt.asarray([0, 1], sycl_queue=q)
        assert str(x) == "[0\n 1]"

    def test_set_precision(self):
        q = get_queue_or_skip()

        dpt.set_print_options(precision=4)
        x = dpt.asarray([1.23450], sycl_queue=q)
        assert str(x) == "[1.2345]"

    def test_threshold_edgeitems(self):
        q = get_queue_or_skip()

        dpt.set_print_options(threshold=1, edgeitems=1)
        x = dpt.arange(9, sycl_queue=q)
        assert str(x) == "[0 ... 8]"
        dpt.set_print_options(edgeitems=9)
        assert str(x) == "[0 1 2 3 4 5 6 7 8]"

    def test_floatmodes(self):
        q = get_queue_or_skip()

        x = dpt.asarray([0.1234, 0.1234678], sycl_queue=q)
        dpt.set_print_options(floatmode="fixed", precision=4)
        assert str(x) == "[0.1234 0.1235]"

        dpt.set_print_options(floatmode="unique")
        assert str(x) == "[0.1234    0.1234678]"

        dpt.set_print_options(floatmode="maxprec")
        assert str(x) == "[0.1234 0.1235]"

        dpt.set_print_options(floatmode="maxprec", precision=8)
        assert str(x) == "[0.1234    0.1234678]"

        dpt.set_print_options(floatmode="maxprec_equal", precision=4)
        assert str(x) == "[0.1234 0.1235]"

        dpt.set_print_options(floatmode="maxprec_equal", precision=8)
        assert str(x) == "[0.1234000 0.1234678]"

    def test_nan_inf_suppress(self):
        q = get_queue_or_skip()

        dpt.set_print_options(nanstr="nan1", infstr="inf1")
        x = dpt.asarray([np.nan, np.inf], sycl_queue=q)
        assert str(x) == "[nan1 inf1]"

    def test_suppress_small(self):
        q = get_queue_or_skip()

        dpt.set_print_options(suppress=True)
        x = dpt.asarray(5e-10, sycl_queue=q)
        assert str(x) == "0."

    def test_sign(self):
        q = get_queue_or_skip()

        x = dpt.asarray([0.0, 1.0, 2.0], sycl_queue=q)
        y = dpt.asarray(1.0, sycl_queue=q)
        z = dpt.asarray([1.0 + 1.0j], sycl_queue=q)
        assert str(x) == "[0. 1. 2.]"
        assert str(y) == "1."
        assert str(z) == "[1.+1.j]"

        dpt.set_print_options(sign="+")
        assert str(x) == "[+0. +1. +2.]"
        assert str(y) == "+1."
        assert str(z) == "[+1.+1.j]"

        dpt.set_print_options(sign=" ")
        assert str(x) == "[ 0.  1.  2.]"
        assert str(y) == " 1."
        assert str(z) == "[ 1.+1.j]"

    def test_numpy(self):
        dpt.set_print_options(numpy=True)
        options = dpt.get_print_options()
        np_options = np.get_printoptions()
        assert all(np_options[k] == options[k] for k in options.keys())


class TestPrintFns(TestPrint):
    @pytest.mark.parametrize(
        "dtype,x_str",
        [
            ("b1", "[False  True  True  True]"),
            ("i1", "[0 1 2 3]"),
            ("u1", "[0 1 2 3]"),
            ("i2", "[0 1 2 3]"),
            ("u2", "[0 1 2 3]"),
            ("i4", "[0 1 2 3]"),
            ("u4", "[0 1 2 3]"),
            ("i8", "[0 1 2 3]"),
            ("u8", "[0 1 2 3]"),
            ("f2", "[0. 1. 2. 3.]"),
            ("f4", "[0. 1. 2. 3.]"),
            ("f8", "[0. 1. 2. 3.]"),
            ("c8", "[0.+0.j 1.+0.j 2.+0.j 3.+0.j]"),
            ("c16", "[0.+0.j 1.+0.j 2.+0.j 3.+0.j]"),
        ],
    )
    def test_print_types(self, dtype, x_str):
        q = get_queue_or_skip()
        skip_if_dtype_not_supported(dtype, q)

        x = dpt.asarray([0, 1, 2, 3], dtype=dtype, sycl_queue=q)
        assert str(x) == x_str

    def test_print_str(self):
        q = get_queue_or_skip()

        x = dpt.asarray(0, sycl_queue=q)
        assert str(x) == "0"

        x = dpt.asarray([np.nan, np.inf], sycl_queue=q)
        assert str(x) == "[nan inf]"

        x = dpt.arange(9, sycl_queue=q)
        assert str(x) == "[0 1 2 3 4 5 6 7 8]"

        y = dpt.reshape(x, (3, 3), copy=True)
        assert str(y) == "[[0 1 2]\n [3 4 5]\n [6 7 8]]"

    def test_print_str_abbreviated(self):
        q = get_queue_or_skip()

        dpt.set_print_options(threshold=0, edgeitems=1)
        x = dpt.arange(9, sycl_queue=q)
        assert str(x) == "[0 ... 8]"

        x = dpt.reshape(x, (3, 3))
        assert str(x) == "[[0 ... 2]\n ...\n [6 ... 8]]"

    def test_usm_ndarray_str_separator(self):
        q = get_queue_or_skip()

        x = dpt.reshape(dpt.arange(4, sycl_queue=q), (2, 2))

        np.testing.assert_equal(
            dpt.usm_ndarray_str(x, prefix="test", separator="   "),
            "[[0   1]\n     [2   3]]",
        )

    def test_print_repr(self):
        q = get_queue_or_skip()

        x = dpt.asarray(0, dtype="int64", sycl_queue=q)
        assert repr(x) == "usm_ndarray(0)"

        x = dpt.asarray([np.nan, np.inf], sycl_queue=q)
        if x.sycl_device.has_aspect_fp64:
            assert repr(x) == "usm_ndarray([nan, inf])"
        else:
            assert repr(x) == "usm_ndarray([nan, inf], dtype=float32)"

        x = dpt.arange(9, sycl_queue=q, dtype="int64")
        assert repr(x) == "usm_ndarray([0, 1, 2, 3, 4, 5, 6, 7, 8])"

        x = dpt.reshape(x, (3, 3))
        np.testing.assert_equal(
            repr(x),
            "usm_ndarray([[0, 1, 2],"
            "\n             [3, 4, 5],"
            "\n             [6, 7, 8]])",
        )

        x = dpt.arange(4, dtype="i4", sycl_queue=q)
        assert repr(x) == "usm_ndarray([0, 1, 2, 3], dtype=int32)"

        dpt.set_print_options(linewidth=1)
        np.testing.assert_equal(
            repr(x),
            "usm_ndarray([0,"
            "\n             1,"
            "\n             2,"
            "\n             3],"
            "\n            dtype=int32)",
        )

    def test_print_repr_abbreviated(self):
        q = get_queue_or_skip()

        dpt.set_print_options(threshold=0, edgeitems=1)
        x = dpt.arange(9, dtype="int64", sycl_queue=q)
        assert repr(x) == "usm_ndarray([0, ..., 8])"

        y = dpt.asarray(x, dtype="i4", copy=True)
        assert repr(y) == "usm_ndarray([0, ..., 8], dtype=int32)"

        x = dpt.reshape(x, (3, 3))
        np.testing.assert_equal(
            repr(x),
            "usm_ndarray([[0, ..., 2],"
            "\n             ...,"
            "\n             [6, ..., 8]])",
        )

        y = dpt.reshape(y, (3, 3))
        np.testing.assert_equal(
            repr(y),
            "usm_ndarray([[0, ..., 2],"
            "\n             ...,"
            "\n             [6, ..., 8]], dtype=int32)",
        )

        dpt.set_print_options(linewidth=1)
        np.testing.assert_equal(
            repr(y),
            "usm_ndarray([[0,"
            "\n              ...,"
            "\n              2],"
            "\n             ...,"
            "\n             [6,"
            "\n              ...,"
            "\n              8]],"
            "\n            dtype=int32)",
        )

    @pytest.mark.parametrize(
        "dtype",
        [
            "i1",
            "u1",
            "i2",
            "u2",
            "i4",
            "u4",
            "u8",
            "f2",
            "f4",
            "c8",
        ],
    )
    def test_repr_appended_dtype(self, dtype):
        q = get_queue_or_skip()
        skip_if_dtype_not_supported(dtype, q)

        x = dpt.empty(4, dtype=dtype)
        assert repr(x).split("=")[-1][:-1] == x.dtype.name

    def test_usm_ndarray_repr_prefix(self):
        q = get_queue_or_skip()

        x = dpt.arange(4, dtype=np.intp, sycl_queue=q)
        np.testing.assert_equal(
            dpt.usm_ndarray_repr(x, prefix="test"), "test([0, 1, 2, 3])"
        )
        x = dpt.reshape(x, (2, 2))
        np.testing.assert_equal(
            dpt.usm_ndarray_repr(x, prefix="test"),
            "test([[0, 1]," "\n      [2, 3]])",
        )


class TestContextManager:
    def test_context_manager_basic(self):
        options = dpt.get_print_options()
        with dpt.print_options(precision=4):
            s = str(dpt.asarray(1.234567))
        assert s == "1.2346"
        assert options == dpt.get_print_options()

    def test_context_manager_as(self):
        with dpt.print_options(precision=4) as x:
            options = x.copy()
        assert options["precision"] == 4



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A collection of utility functions.
"""

from ._compute_follows_data import (
    ExecutionPlacementError,
    get_coerced_usm_type,
    get_execution_queue,
    validate_usm_type,
)
from ._onetrace_context import onetrace_enabled

__all__ = [
    "get_execution_queue",
    "get_coerced_usm_type",
    "validate_usm_type",
    "onetrace_enabled",
    "ExecutionPlacementError",
]



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    **Data Parallel Tensor Collection** is a collection of tensor
    implementations that implement Python data API
    (https://data-apis.github.io/array-api/latest/) standard.

"""

from dpctl.tensor._copy_utils import asnumpy, astype, copy, from_numpy, to_numpy
from dpctl.tensor._ctors import (
    arange,
    asarray,
    empty,
    empty_like,
    eye,
    full,
    full_like,
    linspace,
    meshgrid,
    ones,
    ones_like,
    tril,
    triu,
    zeros,
    zeros_like,
)
from dpctl.tensor._data_types import (
    bool,
    complex64,
    complex128,
    dtype,
    float16,
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)
from dpctl.tensor._device import Device
from dpctl.tensor._dlpack import from_dlpack
from dpctl.tensor._indexing_functions import extract, nonzero, place, put, take
from dpctl.tensor._manipulation_functions import (
    broadcast_arrays,
    broadcast_to,
    can_cast,
    concat,
    expand_dims,
    finfo,
    flip,
    iinfo,
    permute_dims,
    result_type,
    roll,
    squeeze,
    stack,
)
from dpctl.tensor._print import (
    get_print_options,
    print_options,
    set_print_options,
    usm_ndarray_repr,
    usm_ndarray_str,
)
from dpctl.tensor._reshape import reshape
from dpctl.tensor._usmarray import usm_ndarray

from ._constants import e, inf, nan, newaxis, pi

__all__ = [
    "Device",
    "usm_ndarray",
    "arange",
    "asarray",
    "astype",
    "copy",
    "empty",
    "zeros",
    "ones",
    "full",
    "eye",
    "linspace",
    "empty_like",
    "zeros_like",
    "ones_like",
    "full_like",
    "flip",
    "reshape",
    "roll",
    "concat",
    "stack",
    "broadcast_arrays",
    "broadcast_to",
    "expand_dims",
    "permute_dims",
    "squeeze",
    "take",
    "put",
    "extract",
    "place",
    "nonzero",
    "from_numpy",
    "to_numpy",
    "asnumpy",
    "from_dlpack",
    "tril",
    "triu",
    "dtype",
    "bool",
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "float16",
    "float32",
    "float64",
    "complex64",
    "complex128",
    "iinfo",
    "finfo",
    "can_cast",
    "result_type",
    "meshgrid",
    "get_print_options",
    "set_print_options",
    "print_options",
    "usm_ndarray_repr",
    "usm_ndarray_str",
    "newaxis",
    "e",
    "pi",
    "nan",
    "inf",
]



#                      Data Parallel Control (dpctl)
#
# Copyright 2020-2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools
from Cython.Build import build_ext

import dpctl

ext = setuptools.Extension(
    "_cython_api",
    ["_cython_api.pyx"],
    include_dirs=[dpctl.get_include()],
    language="c++",
)

setuptools.setup(
    name="test_cython_api",
    version="0.0.0",
    ext_modules=[ext],
    cmdclass={"build_ext": build_ext},
)
